import json
from copy import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Generator, Iterator, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from aislib.misc_utils import ensure_path_exists
from torch.utils.data import Dataset
from torch.utils.data._utils.collate import default_collate

from eir.data_load.data_preparation_modules.imputation import (
    impute_missing_modalities_wrapper,
)
from eir.data_load.datasets import al_getitem_return
from eir.models.model_training_utils import predict_on_batch
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.torchtext_port.vocab import Vocab
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.schema_modules.output_schemas_sequence import (
    SequenceOutputSamplingConfig,
)
from eir.setup.schemas import OutputConfig, SequenceOutputTypeConfig
from eir.train_utils import utils
from eir.train_utils.evaluation_modules.evaluation_handlers_utils import (
    SpecialTokens,
    convert_model_inputs_to_raw,
    decode_tokens,
    extract_input_types,
    general_pre_process_prepared_inputs,
    get_batch_generator,
    get_dataset_loader_single_sample_generator,
    get_special_tokens,
    post_prepare_manual_inputs,
    prepare_base_input,
    prepare_manual_sample_data,
    remove_special_tokens_from_string,
    serialize_raw_inputs,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.predict import PredictExperiment, PredictHooks
    from eir.serve import ServeExperiment
    from eir.train import Experiment, al_input_objects_as_dict
    from eir.train_utils.step_logic import Hooks

logger = get_logger(name=__name__)


@dataclass
class SequenceOutputEvalSamples:
    auto_samples: Dict[str, list["SequenceOutputEvalSample"]]
    manual_samples: Dict[str, list["SequenceOutputEvalSample"]]


@dataclass()
class SequenceOutputEvalSample:
    inputs_to_model: Dict[str, Any]
    target_labels: Dict[str, Any]
    sample_id: str


def sequence_out_single_sample_evaluation_wrapper(
    experiment: Union["Experiment", "PredictExperiment", "ServeExperiment"],
    input_objects: "al_input_objects_as_dict",
    auto_dataset_to_load_from: Dataset,
    iteration: int,
    output_folder: str,
) -> None:

    gc = experiment.configs.global_config

    default_eir_hooks = experiment.hooks
    assert default_eir_hooks is not None

    output_configs = experiment.configs.output_configs

    if not any(i.sampling_config for i in output_configs):
        return

    manual_samples = get_sequence_output_manual_input_samples(
        output_configs=output_configs,
        input_objects=input_objects,
    )

    auto_validation_generator = get_dataset_loader_single_sample_generator(
        dataset=auto_dataset_to_load_from
    )
    auto_samples = get_sequence_output_auto_validation_samples(
        output_configs=output_configs,
        input_objects=input_objects,
        eval_sample_iterator=auto_validation_generator,
    )
    eval_samples_base = SequenceOutputEvalSamples(
        auto_samples=auto_samples,
        manual_samples=manual_samples,
    )

    for config in output_configs:
        if config.output_info.output_type != "sequence":
            continue

        cur_input_name = config.output_info.output_name
        cur_output_name = config.output_info.output_name

        not_in_manual_samples = cur_output_name not in eval_samples_base.manual_samples
        not_in_auto_samples = cur_output_name not in eval_samples_base.auto_samples
        if not_in_manual_samples and not_in_auto_samples:
            continue

        cur_sample_output_folder = utils.prepare_sample_output_folder(
            output_folder=output_folder,
            output_name=cur_output_name,
            column_name=cur_input_name,
            iteration=iteration,
        )

        output_type_info = config.output_type_info
        assert isinstance(output_type_info, SequenceOutputTypeConfig)

        assert config.sampling_config is not None

        if output_type_info.sequence_operation == "autoregressive":
            sample_generator = _get_eval_sample_generator(
                eval_samples=eval_samples_base,
                output_name=cur_output_name,
            )

            batch_generator = get_batch_generator(
                iterator=enumerate(sample_generator),
                batch_size=gc.be.batch_size,
            )

            assert isinstance(config.sampling_config, SequenceOutputSamplingConfig)

            meta = {}

            for batch in batch_generator:
                batch_indices, batch_eval_data = zip(*batch)
                batch_eval_types, batch_eval_samples = zip(*batch_eval_data)

                batch_generated_tokens = autoregressive_sequence_generation(
                    input_objects=input_objects,
                    eval_samples=batch_eval_samples,
                    seq_output_name=cur_output_name,
                    experiment=experiment,
                    default_eir_hooks=default_eir_hooks,
                    sampling_config=config.sampling_config,
                )

                cur_input_object = input_objects[cur_input_name]
                assert isinstance(cur_input_object, ComputedSequenceInputInfo)
                assert cur_input_object.tokenizer is not None

                for eval_type, idx, eval_sample, generated_tokens in zip(
                    batch_eval_types,
                    batch_indices,
                    batch_eval_samples,
                    batch_generated_tokens,
                ):

                    generated_sample = decode_tokens(
                        tokens=generated_tokens,
                        vocab=cur_input_object.vocab,
                        split_on=output_type_info.split_on,
                    )
                    special_tokens = get_special_tokens(
                        tokenizer=cur_input_object.tokenizer,
                        vocab=cur_input_object.vocab,
                    )
                    generated_sample = remove_special_tokens_from_string(
                        string=generated_sample,
                        special_tokens=special_tokens,
                    )

                    cur_output_path = (
                        cur_sample_output_folder / eval_type / f"{idx}_generated.txt"
                    )
                    ensure_path_exists(path=cur_output_path)
                    cur_output_path.write_text(data=generated_sample)

                    cur_inputs_output_path = (
                        cur_sample_output_folder / eval_type / f"{idx}_inputs"
                    )

                    raw_inputs = convert_model_inputs_to_raw(
                        inputs_to_model=eval_sample.inputs_to_model,
                        input_objects=input_objects,
                    )

                    serialize_raw_inputs(
                        raw_inputs=raw_inputs,
                        input_objects=input_objects,
                        output_path=cur_inputs_output_path,
                    )

                    cur_id = eval_sample.sample_id
                    meta[cur_id] = {
                        "generated": str(cur_output_path.relative_to(output_folder)),
                        "inputs": str(
                            cur_inputs_output_path.relative_to(output_folder)
                        ),
                        "index": idx,
                    }

                meta_path = cur_sample_output_folder / "meta.json"
                with open(meta_path, "w") as f:
                    json.dump(meta, f, indent=4)


def get_sequence_output_manual_input_samples(
    output_configs: Sequence[OutputConfig],
    input_objects: "al_input_objects_as_dict",
) -> Dict[str, list[SequenceOutputEvalSample]]:
    prepared_samples: dict[str, list[SequenceOutputEvalSample]] = {}

    for config_idx, config in enumerate(output_configs):
        if not config.sampling_config or config.output_info.output_type != "sequence":
            continue

        assert isinstance(config.sampling_config, SequenceOutputSamplingConfig)

        sample_data_from_yaml = config.sampling_config.manual_inputs
        output_name = config.output_info.output_name

        prepared_samples[output_name] = []

        for idx, single_sample_inputs in enumerate(sample_data_from_yaml):
            prepared_inputs = prepare_manual_sample_data(
                sample_inputs=single_sample_inputs,
                input_objects=input_objects,
            )

            imputed_inputs = impute_missing_modalities_wrapper(
                inputs_values=prepared_inputs,
                inputs_objects=input_objects,
            )

            final_inputs = post_prepare_manual_inputs(
                prepared_inputs=imputed_inputs,
                output_name=output_name,
                input_objects=input_objects,
            )

            cur_eval_sample = SequenceOutputEvalSample(
                inputs_to_model=final_inputs,
                target_labels={},
                sample_id=f"manual_{idx}",
            )

            prepared_samples[output_name].append(cur_eval_sample)

    return prepared_samples


def get_sequence_output_auto_validation_samples(
    output_configs: Sequence[OutputConfig],
    input_objects: "al_input_objects_as_dict",
    eval_sample_iterator: Iterator[al_getitem_return],
) -> Dict[str, list[SequenceOutputEvalSample]]:
    prepared_eval_samples: dict[str, list[SequenceOutputEvalSample]] = {}
    input_types = extract_input_types(input_objects=input_objects)

    for config_idx, config in enumerate(output_configs):
        if not config.sampling_config or config.output_info.output_type != "sequence":
            continue

        assert isinstance(config.sampling_config, SequenceOutputSamplingConfig)

        output_name = config.output_info.output_name

        prepared_eval_samples[output_name] = []

        for i in range(config.sampling_config.n_eval_inputs):
            input_to_model, target_labels, cur_id = next(eval_sample_iterator)

            cur_inputs_masked = _mask_targets_for_auto_eval_generation(
                inputs=input_to_model,
                output_name=output_name,
                input_types=input_types,
            )

            cur_eval_sample = SequenceOutputEvalSample(
                inputs_to_model=cur_inputs_masked,
                target_labels=target_labels,
                sample_id=cur_id[0],
            )

            prepared_eval_samples[output_name].append(cur_eval_sample)

    return prepared_eval_samples


def _mask_targets_for_auto_eval_generation(
    inputs: Dict[str, Any],
    output_name: str,
    input_types: Dict[str, str],
) -> Dict[str, Any]:
    raw_inputs_masked = {}

    for input_name, raw_input in inputs.items():
        if input_name == output_name:
            if input_types[input_name] == "sequence":
                raw_inputs_masked[output_name] = torch.tensor([], dtype=torch.long)
            else:
                raise NotImplementedError()

        else:
            raw_inputs_masked[input_name] = raw_input

    return raw_inputs_masked


def _get_eval_sample_generator(
    eval_samples: SequenceOutputEvalSamples, output_name: str
) -> Generator[Tuple[str, SequenceOutputEvalSample], None, None]:
    cur_config_auto_samples = eval_samples.auto_samples[output_name]
    cur_config_manual_samples = eval_samples.manual_samples[output_name]

    for eval_sample in cur_config_auto_samples:
        yield "auto", eval_sample

    for eval_sample in cur_config_manual_samples:
        yield "manual", eval_sample


@torch.inference_mode()
def autoregressive_sequence_generation(
    input_objects: "al_input_objects_as_dict",
    eval_samples: tuple[SequenceOutputEvalSample, ...],
    seq_output_name: str,
    experiment: Union["Experiment", "PredictExperiment", "ServeExperiment"],
    default_eir_hooks: Union["Hooks", "PredictHooks"],
    sampling_config: SequenceOutputSamplingConfig,
) -> list[list[int]]:
    """
    Note that it's currently a bit weird / perhaps suboptimal how we are dealing
    with the input data here. While normally one could have the final, prepared
    input (including embeddings) as a direct input to the model, we have here
    what is more similar to what is returned from the data loader, namely e.g.
    token IDs. The sequence of IDs is then used in the loop and appended to,
    which in turn goes through the general batch preparation hook, which e.g.
    looks up the embeddings.
    """
    output_object = experiment.outputs[seq_output_name]
    input_object = input_objects[seq_output_name]
    max_length = sampling_config.generated_sequence_length

    assert isinstance(output_object, ComputedSequenceOutputInfo)
    assert isinstance(input_object, ComputedSequenceInputInfo)

    _check_vocab_consistency(
        output_object_vocab=output_object.vocab,
        input_object_vocab=input_object.vocab,
    )

    assert output_object.tokenizer is not None
    st = get_special_tokens(
        tokenizer=output_object.tokenizer,
        vocab=output_object.vocab,
    )

    autoregressive_pre_batch = prepare_autoregressive_sampling_batch(
        eval_samples=eval_samples,
        seq_output_name=seq_output_name,
        special_tokens=st,
    )
    prepared_sample_inputs = autoregressive_pre_batch.prepared_inputs
    prepared_targets = autoregressive_pre_batch.prepared_targets
    generated_tokens = autoregressive_pre_batch.generated_tokens
    indices = autoregressive_pre_batch.indices

    indices_has_finished = {}
    for i in range(0, sampling_config.generated_sequence_length):

        target_indices = []
        for sample_index, _ in enumerate(eval_samples):
            cur_generated_tokens = generated_tokens[sample_index]

            cur_prepared_sample_inputs = prepared_sample_inputs[sample_index]
            cur_prepared_sample_inputs = _prepare_current_autoregressive_input(
                prepared_sample_inputs=cur_prepared_sample_inputs,
                generated_tokens=cur_generated_tokens,
                seq_output_name=seq_output_name,
                max_length=output_object.computed_max_length,
                pad_idx=st.pad_idx,
            )

            cur_target_index = _compute_target_index(
                current_generated_length=indices[sample_index],
                max_length=output_object.computed_max_length,
            )

            target_indices.append(cur_target_index)
            prepared_sample_inputs[sample_index] = cur_prepared_sample_inputs
            indices[sample_index] += 1

        all_inputs = default_collate(prepared_sample_inputs)
        all_targets = default_collate(prepared_targets)
        all_ids = [eval_sample.sample_id for eval_sample in eval_samples]

        batch = general_pre_process_prepared_inputs(
            prepared_inputs=all_inputs,
            target_labels=all_targets,
            sample_ids=all_ids,
            experiment=experiment,
            custom_hooks=default_eir_hooks,
        )

        outputs = predict_on_batch(
            model=experiment.model,
            inputs=batch.inputs,
        )

        next_token_indices = sample_next_token_index_from_output(
            outputs=outputs,
            seq_output_name=seq_output_name,
            sampling_config=sampling_config,
            current_target_indices=target_indices,
        )

        for sample_index, _ in enumerate(eval_samples):
            if sample_index in indices_has_finished:
                continue

            next_token_index = next_token_indices[sample_index]

            is_eos = next_token_index == st.eos_idx
            is_at_max_length = indices[sample_index] == max_length

            if is_eos or is_at_max_length:
                indices_has_finished[sample_index] = True
                continue

            generated_tokens[sample_index].append(next_token_index)

    return generated_tokens


@dataclass(frozen=False)
class AutoRegressiveSamplingBatch:
    prepared_inputs: list[dict[str, torch.Tensor]]
    prepared_targets: list[dict[str, torch.Tensor]]
    indices: list[int]
    generated_tokens: list[list[int]]


def prepare_autoregressive_sampling_batch(
    eval_samples: Sequence[SequenceOutputEvalSample],
    seq_output_name: str,
    special_tokens: SpecialTokens,
) -> AutoRegressiveSamplingBatch:
    prepared_sample_inputs = []
    prepared_targets = []
    generated_tokens = []
    indices = []

    for eval_sample in eval_samples:
        cur_inputs = eval_sample.inputs_to_model
        cur_prepared = prepare_base_input(prepared_inputs=cur_inputs)
        prepared_sample_inputs.append(cur_prepared)
        prepared_targets.append(eval_sample.target_labels)

        cur_generated_tokens_base = _extract_base_generated_tokens(
            prepared_inputs=cur_prepared,
            seq_output_name=seq_output_name,
        )

        cur_generated_tokens = _ensure_no_extra_padding(
            tokens=cur_generated_tokens_base,
            pad_idx=special_tokens.pad_idx,
        )

        generated_tokens.append(cur_generated_tokens)
        indices.append(len(cur_generated_tokens))

    return AutoRegressiveSamplingBatch(
        prepared_inputs=prepared_sample_inputs,
        prepared_targets=prepared_targets,
        indices=indices,
        generated_tokens=generated_tokens,
    )


def _check_vocab_consistency(
    output_object_vocab: Vocab, input_object_vocab: Vocab
) -> None:
    assert output_object_vocab.get_stoi() == input_object_vocab.get_stoi()
    assert output_object_vocab.get_itos() == input_object_vocab.get_itos()


def _extract_base_generated_tokens(
    prepared_inputs: Dict[str, torch.Tensor],
    seq_output_name: str,
) -> list[int]:
    """
    Note that we are expecting / enforcing the token IDs being passed in here,
    not the embeddings.
    """
    tensor_base = prepared_inputs[seq_output_name]

    assert tensor_base.dim() == 1, (tensor_base, tensor_base.dim())

    list_base = tensor_base.tolist()
    assert isinstance(list_base, list)

    return list_base


def _ensure_no_extra_padding(tokens: list[int], pad_idx: int) -> list[int]:
    """
    Needed in case the sequence is already at max due to padding, as the autoregressive
    functionality itself takes care of shifting, bos insertion, padding etc. So, it
    expects just the sequence as given, without any padding. However, some processing
    code might have added padding to the end, so we have this little safeguard here.
    """
    generated_tokens = [i for i in tokens if i != pad_idx]
    if generated_tokens:
        assert generated_tokens[-1] != pad_idx

    return generated_tokens


def _compute_target_index(current_generated_length: int, max_length: int) -> int:
    if current_generated_length < max_length:
        return current_generated_length
    else:
        return max_length - 1


def _prepare_current_autoregressive_input(
    prepared_sample_inputs: dict[str, Any],
    generated_tokens: list[int],
    seq_output_name: str,
    max_length: int,
    pad_idx: int,
) -> dict[str, torch.Tensor]:
    """
    Assuming max_length of 5:

    Truncate (from start):
        - [A, B, C, D, E] -> [B, C, D, E]
    Pad:
        - [B, C, D, E] -> [B, C, D, E, pad]
    In autoregressive batch preparation hook:
        - [B, C, D, E, pad] -> [bos, B, C, D, E, pad] -> [bos, B, C, D, E]

    So essentially the pad is a placeholder to that we need here due to how
    the batch is later (in the batch preparation hook) prepared, namely
    as it inserts the bos at the beginning and cuts off at the end, so in the case
    where the sequence is already at max_length, we need to pad it to make sure
    that the latest token is preserved.

    The reason for padding with the BOS token is that the tensor we get here
    is already at max_length. If we had e.g. a full, long sequence, we could
    simply slice that directly (+1 for the target), but here we need to pad
    the input at the beginning, opting for a BOS token.
    """
    current_sequence = torch.tensor(generated_tokens, dtype=torch.long)
    current_sequence = _maybe_truncate_autoregressive_sequence(
        current_sequence=current_sequence,
        max_length=max_length,
    )

    assert current_sequence.dim() == 1, current_sequence.shape

    pad_size = max_length - len(current_sequence)
    assert pad_size >= 1, pad_size

    current_sequence = F.pad(current_sequence, (0, pad_size), value=pad_idx)
    assert current_sequence.dim() == 1, current_sequence.shape

    current_inputs_copy = copy(prepared_sample_inputs)
    current_inputs_copy[seq_output_name] = current_sequence

    return current_inputs_copy


def _maybe_truncate_autoregressive_sequence(
    current_sequence: torch.Tensor,
    max_length: int,
) -> torch.Tensor:
    """
    +1 to account for bos token being inserted and last token removed later.
    If already at max length, we truncate from the start and leave 1 element
    missing (+1) to allow for padding.
    """
    sequence_length = len(current_sequence)

    if sequence_length >= max_length:
        start = sequence_length - max_length + 1
        current_sequence = current_sequence[start:]

    return current_sequence


def sample_next_token_index_from_output(
    outputs: dict[str, dict[str, torch.Tensor]],
    seq_output_name: str,
    sampling_config: SequenceOutputSamplingConfig,
    current_target_indices: list[int],
) -> list[int]:
    cur_logits = outputs[seq_output_name][seq_output_name]
    batch_indices = torch.arange(cur_logits.size(0))
    cur_position_logits = cur_logits[batch_indices, current_target_indices, :]

    filtered_logits = top_k_top_p_filtering(
        logits=cur_position_logits,
        top_k=sampling_config.top_k,
        top_p=sampling_config.top_p,
    )
    probabilities = F.softmax(input=filtered_logits, dim=-1)
    next_token_indices = torch.multinomial(input=probabilities, num_samples=1)
    next_token_indices_list = next_token_indices.squeeze(1).tolist()

    return next_token_indices_list


def top_k_top_p_filtering(
    logits: torch.Tensor,
    top_k: int = 0,
    top_p: float = 0.0,
    filter_value: float = -float("Inf"),
) -> torch.Tensor:
    assert (
        logits.dim() == 2
    ), f"Expected 2D tensor (batch, vocab_scores). Got: {logits.dim()}D."

    batch_size, vocab_size = logits.size()

    top_k = min(top_k, logits.size(-1))
    if top_k > 0:
        indices_to_remove = logits < torch.topk(logits, top_k)[0][..., -1, None]
        logits[indices_to_remove] = filter_value

    if top_p > 0.0:
        sorted_logits, sorted_indices = torch.sort(logits, descending=True)
        cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

        sorted_indices_to_remove = cumulative_probs > top_p
        sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
        sorted_indices_to_remove[..., 0] = 0

        for i in range(batch_size):
            indices_to_remove = sorted_indices[i, sorted_indices_to_remove[i]]
            logits[i, indices_to_remove] = filter_value

    return logits



import warnings
from typing import Literal

import numpy as np
import torch
from torch.nn.functional import pad

from eir.data_load.data_source_modules import deeplake_ops


def _load_deeplake_sample(
    data_pointer: int,
    input_source: str,
    inner_key: str,
) -> np.ndarray | str:
    """
    Deeplake warns about indexing directly into a DS, vs. random access. For now, we'll
    use this random access pattern here as we have to be able to connect to other
    data sources (which might be outside deeplake).
    """
    assert inner_key is not None
    deeplake_ds = deeplake_ops.load_deeplake_dataset(data_source=input_source)
    deeplake_ds_index = data_pointer
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        sample_data = deeplake_ds[deeplake_ds_index][inner_key]

    return sample_data


def process_tensor_to_length(
    tensor: torch.Tensor,
    max_length: int,
    sampling_strategy_if_longer: Literal["from_start", "uniform"],
    padding_value: int = 0,
) -> torch.Tensor:
    tensor_length = len(tensor)

    if tensor_length > max_length:
        if sampling_strategy_if_longer == "from_start":
            truncated_tensor = tensor[:max_length]
            return truncated_tensor

        if sampling_strategy_if_longer == "uniform":
            uniformly_sampled_tensor = _sample_sequence_uniform(
                tensor=tensor, tensor_length=tensor_length, max_length=max_length
            )
            return uniformly_sampled_tensor

    right_padding = max_length - tensor_length
    padded_tensor = pad(input=tensor, pad=[0, right_padding], value=padding_value)

    return padded_tensor


def _sample_sequence_uniform(
    tensor: torch.Tensor, tensor_length: int, max_length: int
) -> torch.Tensor:
    random_index_start = torch.randperm(max(1, tensor_length - max_length))[0]
    random_index_end = random_index_start + max_length
    return tensor[random_index_start:random_index_end]



from typing import Optional

from eir.setup.schema_modules.tensor_broker_schemas import (
    TensorBrokerConfig,
    TensorMessageConfig,
)


def set_up_tensor_broker_config(
    tensor_broker_config: dict,
) -> Optional[TensorBrokerConfig]:
    if not tensor_broker_config:
        return None

    tensor_message_configs = []
    for message_config in tensor_broker_config["message_configs"]:
        tensor_message_configs.append(TensorMessageConfig(**message_config))

    return TensorBrokerConfig(message_configs=tensor_message_configs)



import math
from collections import abc
from dataclasses import dataclass
from typing import Any, Callable, Dict, Literal, Optional, Tuple, Union

import torch
from torch import nn
from transformers import AutoConfig, AutoModel, PreTrainedModel

from eir.models.input.sequence.sequence_models import SequenceModelClassGetterFunction
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
    SequenceModelConfig,
    TransformerFeatureExtractor,
    TransformerWrapperModel,
    get_embedding_dim_for_sequence_model,
)
from eir.models.models_utils import get_output_dimensions_for_input
from eir.setup.setup_utils import get_unsupported_hf_models
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


@dataclass
class SequenceModelObjectsForWrapperModel:
    feature_extractor: Union[TransformerFeatureExtractor, nn.Module]
    embeddings: Optional[nn.Embedding]
    embedding_dim: int
    external: bool
    known_out_features: Union[None, int]


def get_sequence_model(
    sequence_model_config: SequenceModelConfig,
    model_registry_lookup: SequenceModelClassGetterFunction,
    num_tokens: int,
    max_length: int,
    embedding_dim: int,
    device: str,
) -> TransformerWrapperModel:
    feature_extractor_max_length, num_chunks = _get_windowed_sequence_parameters(
        max_length=max_length,
        window_size=sequence_model_config.window_size,
    )

    objects_for_wrapper = _get_sequence_feature_extractor_objects_for_wrapper_model(
        model_type=sequence_model_config.model_type,
        model_registry_lookup=model_registry_lookup,
        pretrained=sequence_model_config.pretrained_model,
        pretrained_frozen=sequence_model_config.freeze_pretrained_model,
        model_config=sequence_model_config.model_init_config,
        num_tokens=num_tokens,
        embedding_dim=embedding_dim,
        feature_extractor_max_length=feature_extractor_max_length,
        num_chunks=num_chunks,
        pool=sequence_model_config.pool,
    )

    device_torch = torch.device(device)
    sequence_model = TransformerWrapperModel(
        feature_extractor=objects_for_wrapper.feature_extractor,
        external_feature_extractor=objects_for_wrapper.external,
        model_config=sequence_model_config,
        embedding_dim=objects_for_wrapper.embedding_dim,
        num_tokens=num_tokens,
        max_length=max_length,
        embeddings=objects_for_wrapper.embeddings,
        device=device,
        pre_computed_num_out_features=objects_for_wrapper.known_out_features,
    ).to(device=device_torch)

    return sequence_model


def _get_windowed_sequence_parameters(
    max_length: int, window_size: int
) -> Tuple[int, int]:
    feature_extractor_max_length = max_length
    num_chunks = 1
    if window_size:
        logger.info(
            "Using sliding model for sequence input as window size was set to %d.",
            window_size,
        )
        feature_extractor_max_length = window_size
        num_chunks = math.ceil(max_length / window_size)

    return feature_extractor_max_length, num_chunks


def _get_sequence_feature_extractor_objects_for_wrapper_model(
    model_type: str,
    model_registry_lookup: SequenceModelClassGetterFunction,
    pretrained: bool,
    pretrained_frozen: bool,
    model_config: Union[BasicTransformerFeatureExtractorModelConfig, dict],
    num_tokens: int,
    embedding_dim: int,
    feature_extractor_max_length: int,
    num_chunks: int,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> SequenceModelObjectsForWrapperModel:
    if "sequence-default" in model_type or model_type.startswith("eir-"):
        model_class = model_registry_lookup(model_type=model_type)
        assert isinstance(
            model_config, BasicTransformerFeatureExtractorModelConfig
        ), model_config
        objects_for_wrapper = _get_basic_sequence_feature_extractor_objects(
            model_config=model_config,
            num_tokens=num_tokens,
            feature_extractor_max_length=feature_extractor_max_length,
            embedding_dim=embedding_dim,
            feature_extractor_class=model_class,
        )
    elif pretrained:
        assert isinstance(model_config, dict), model_config
        objects_for_wrapper = _get_pretrained_hf_sequence_feature_extractor_objects(
            model_name=model_type,
            frozen=pretrained_frozen,
            feature_extractor_max_length=feature_extractor_max_length,
            num_chunks=num_chunks,
            num_tokens=num_tokens,
            pool=pool,
        )
    else:
        assert isinstance(model_config, dict), model_config
        objects_for_wrapper = _get_hf_sequence_feature_extractor_objects(
            model_name=model_type,
            model_config=model_config,
            feature_extractor_max_length=feature_extractor_max_length,
            num_chunks=num_chunks,
            pool=pool,
        )

    return objects_for_wrapper


def _get_manual_out_features_for_external_feature_extractor(
    input_length: int,
    embedding_dim: int,
    num_chunks: int,
    feature_extractor: nn.Module,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> int:
    input_shape = _get_sequence_input_dim(
        input_length=input_length,
        embedding_dim=embedding_dim,
    )
    out_feature_shape = get_output_dimensions_for_input(
        module=feature_extractor,
        input_shape=input_shape,
        hf_model=True,
        pool=pool,
    )
    manual_out_features = out_feature_shape.numel() * num_chunks

    return manual_out_features


def _get_sequence_input_dim(
    input_length: int, embedding_dim: int
) -> Tuple[int, int, int]:
    return 1, input_length, embedding_dim


def _get_pretrained_hf_sequence_feature_extractor_objects(
    model_name: str,
    num_tokens: int,
    frozen: bool,
    feature_extractor_max_length: int,
    num_chunks: int,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> SequenceModelObjectsForWrapperModel:
    _warn_about_unsupported_hf_model(model_name=model_name)

    pretrained_model = _get_hf_pretrained_model(model_name=model_name)
    pretrained_model.resize_token_embeddings(new_num_tokens=num_tokens)
    pretrained_model_embeddings = pretrained_model.get_input_embeddings()
    feature_extractor = pretrained_model

    if frozen:
        logger.info("Freezing weights and embeddings of model '%s'.", model_name)
        for param in feature_extractor.parameters():
            param.requires_grad = False
        for param in pretrained_model_embeddings.parameters():
            param.requires_grad = False

    pretrained_embedding_dim = _pretrained_hf_model_embedding_dim(
        embeddings=pretrained_model_embeddings,
        model_name=model_name,
    )
    known_out_features = _get_manual_out_features_for_external_feature_extractor(
        input_length=feature_extractor_max_length,
        embedding_dim=pretrained_embedding_dim,
        num_chunks=num_chunks,
        feature_extractor=feature_extractor,
        pool=pool,
    )
    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=pretrained_model,
        embeddings=pretrained_model_embeddings,
        embedding_dim=pretrained_embedding_dim,
        external=True,
        known_out_features=known_out_features,
    )

    return objects_for_wrapper


def _get_hf_pretrained_model(model_name: str) -> PreTrainedModel:
    pretrained_model = AutoModel.from_pretrained(
        pretrained_model_name_or_path=model_name
    )
    logger.info(
        "Loaded external pre-trained model '%s'. Note that this means that "
        "many configurations that might be set in input_type_info and model_config"
        "(e.g. 'embedding_dim') have no effect, as the default settings for "
        "the pre-trained model are used.",
        model_name,
    )
    return pretrained_model


def _get_hf_sequence_feature_extractor_objects(
    model_name: str,
    model_config: Dict[str, Any],
    feature_extractor_max_length: int,
    num_chunks: int,
    pool: Union[Literal["max"], Literal["avg"], None],
) -> SequenceModelObjectsForWrapperModel:
    _warn_about_unsupported_hf_model(model_name=model_name)

    feature_extractor = _get_hf_model(model_name=model_name, model_config=model_config)
    pretrained_model_embeddings = feature_extractor.get_input_embeddings()

    pretrained_embedding_dim = _pretrained_hf_model_embedding_dim(
        embeddings=pretrained_model_embeddings,
        model_name=model_name,
    )
    known_out_features = _get_manual_out_features_for_external_feature_extractor(
        input_length=feature_extractor_max_length,
        embedding_dim=pretrained_embedding_dim,
        num_chunks=num_chunks,
        feature_extractor=feature_extractor,
        pool=pool,
    )
    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=feature_extractor,
        embeddings=None,
        embedding_dim=pretrained_embedding_dim,
        external=True,
        known_out_features=known_out_features,
    )

    return objects_for_wrapper


def _pretrained_hf_model_embedding_dim(
    embeddings: nn.Module, model_name: str = ""
) -> int:
    if hasattr(embeddings, "embedding_dim"):
        emb_dim = embeddings.embedding_dim
    elif hasattr(embeddings, "dim"):
        emb_dim = embeddings.dim
    elif hasattr(embeddings, "emb_layers"):
        layers = embeddings.emb_layers
        assert isinstance(layers, (abc.Sequence, nn.ModuleList))
        emb_dim = layers[0].embedding_dim
    else:
        raise ValueError(f"Could not find embedding dimension for model {model_name}.")

    assert isinstance(emb_dim, int)
    return emb_dim


def _get_hf_model(model_name: str, model_config: Dict[str, Any]) -> PreTrainedModel:
    config = AutoConfig.for_model(model_type=model_name, **model_config)
    model = AutoModel.from_config(config=config)
    logger.info(
        "Set up external (not using pre-trained weights) model '%s'. "
        "With configuration %s. Note that setting up external models ignores values "
        "for fields 'embedding_dim' and 'max_length' from input_type_info "
        "configuration. To configure these models, set the relevant values in the "
        "model_config field of the input configuration.",
        model_name,
        config,
    )
    return model


def _get_basic_sequence_feature_extractor_objects(
    model_config: BasicTransformerFeatureExtractorModelConfig,
    num_tokens: int,
    feature_extractor_max_length: int,
    embedding_dim: int,
    feature_extractor_class: Callable = TransformerFeatureExtractor,
) -> SequenceModelObjectsForWrapperModel:
    parsed_embedding_dim = get_embedding_dim_for_sequence_model(
        embedding_dim=embedding_dim,
        num_tokens=num_tokens,
        num_heads=model_config.num_heads,
    )

    feature_extractor = feature_extractor_class(
        model_config=model_config,
        num_tokens=num_tokens,
        max_length=feature_extractor_max_length,
        embedding_dim=parsed_embedding_dim,
    )

    objects_for_wrapper = SequenceModelObjectsForWrapperModel(
        feature_extractor=feature_extractor,
        embeddings=None,
        embedding_dim=parsed_embedding_dim,
        external=False,
        known_out_features=0,
    )

    return objects_for_wrapper


def _warn_about_unsupported_hf_model(model_name: str) -> None:
    unsupported_models = get_unsupported_hf_models()
    if model_name in unsupported_models.keys():
        reason = unsupported_models[model_name]
        logger.warning(
            "Model '%s' has not been tested for compatibility with EIR due to "
            "reason: '%s'. It is very likely that it will not work straight out of "
            "the box with EIR.",
            model_name,
            reason,
        )



from dataclasses import dataclass, field
from typing import Any, Dict, Literal, Optional, Protocol, Tuple, Type, Union, cast

from torch import nn

from eir.models.fusion import fusion
from eir.models.input.array.array_models import CNNModelConfig
from eir.models.input.image.image_models import get_image_model_class
from eir.models.input.sequence.sequence_models import get_sequence_model_class
from eir.models.input.tabular.tabular import get_unique_values_from_transformers
from eir.models.meta import meta
from eir.models.meta.meta_utils import (
    al_fusion_modules,
    al_input_modules,
    al_output_modules,
)
from eir.models.model_setup_modules.input_model_setup.input_model_setup_array import (
    get_array_input_feature_extractor,
    get_array_model,
)
from eir.models.model_setup_modules.input_model_setup.input_model_setup_image import (
    get_image_model,
)
from eir.models.model_setup_modules.input_model_setup.input_model_setup_omics import (
    get_omics_model_from_model_config,
)
from eir.models.model_setup_modules.input_model_setup.input_model_setup_sequence import (  # noqa
    get_sequence_model,
)
from eir.models.model_setup_modules.input_model_setup.input_model_setup_tabular import (
    SimpleTabularModel,
    get_tabular_model,
)
from eir.models.model_setup_modules.output_model_setup_modules.output_model_setup_array import (  # noqa
    get_array_or_image_output_module_from_model_config,
)
from eir.models.model_setup_modules.output_model_setup_modules.output_model_setup_sequence import (  # noqa
    get_sequence_output_module_from_model_config,
)
from eir.models.model_setup_modules.output_model_setup_modules.output_model_setup_tabular import (  # noqa
    get_tabular_output_module_from_model_config,
)
from eir.models.output.sequence.sequence_output_modules import (
    SequenceOutputModuleConfig,
)
from eir.models.tensor_broker.tensor_broker import get_tensor_broker
from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup import schemas
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.common import DataDimensions
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import ImageInputDataConfig, TabularInputDataConfig
from eir.train_utils.distributed import AttrDelegatedDistributedDataParallel
from eir.train_utils.optim import AttrDelegatedSWAWrapper

al_meta_model = Union[
    meta.MetaModel, AttrDelegatedDistributedDataParallel, AttrDelegatedSWAWrapper
]


al_data_dimensions = Dict[
    str,
    Union[
        DataDimensions,
        "OmicsDataDimensions",
        "SequenceDataDimensions",
    ],
]


@dataclass
class SequenceDataDimensions(DataDimensions):
    @property
    def max_length(self) -> int:
        return self.height

    @property
    def embedding_dim(self) -> int:
        return self.width


@dataclass
class OmicsDataDimensions(DataDimensions):
    @property
    def num_snps(self) -> int:
        return self.width

    @property
    def one_hot_encoding_dim(self) -> int:
        return self.height


def get_default_meta_class(
    meta_model_type: str,
) -> Type[al_meta_model]:
    if meta_model_type == "default":
        return meta.MetaModel
    raise ValueError(f"Unrecognized meta model type: {meta_model_type}.")


class MetaClassGetterCallable(Protocol):
    def __call__(
        self,
        meta_model_type: str,
    ) -> Type[al_meta_model]: ...


def get_meta_model_class_and_kwargs_from_configs(
    global_config: schemas.GlobalConfig,
    fusion_config: schemas.FusionConfig,
    inputs_as_dict: al_input_objects_as_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    meta_class_getter: MetaClassGetterCallable = get_default_meta_class,
    strict: bool = True,
) -> Tuple[Type[al_meta_model], Dict[str, Any]]:
    meta_model_class = meta_class_getter(meta_model_type="default")

    meta_model_kwargs = get_meta_model_kwargs_from_configs(
        global_config=global_config,
        fusion_config=fusion_config,
        inputs_as_dict=inputs_as_dict,
        outputs_as_dict=outputs_as_dict,
        strict=strict,
    )

    return meta_model_class, meta_model_kwargs


def get_meta_model_kwargs_from_configs(
    global_config: schemas.GlobalConfig,
    fusion_config: schemas.FusionConfig,
    inputs_as_dict: al_input_objects_as_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    strict: bool = True,
) -> Dict[str, Any]:

    kwargs: dict[str, Any] = {}
    input_modules = get_input_modules(
        inputs_as_dict=inputs_as_dict,
        device=global_config.be.device,
    )
    kwargs["input_modules"] = input_modules

    out_feature_per_feature_extractor = _get_feature_extractors_num_output_features(
        input_modules=input_modules
    )
    output_types = _get_output_types(outputs_as_dict=outputs_as_dict)
    diffusion_targets = _extract_diffusion_targets(outputs_as_dict=outputs_as_dict)

    fusion_modules = fusion.get_fusion_modules(
        fusion_model_type=fusion_config.model_type,
        model_config=fusion_config.model_config,
        modules_to_fuse=input_modules,
        out_feature_per_feature_extractor=out_feature_per_feature_extractor,
        output_types=output_types,
        any_diffusion=diffusion_targets != {},
        strict=strict,
    )
    kwargs["fusion_modules"] = fusion_modules

    computed_out_dimension = _get_maybe_computed_out_dims(fusion_modules=fusion_modules)
    feature_dims_and_types = get_all_feature_extractor_dimensions_and_types(
        inputs_as_dict=inputs_as_dict,
        input_modules=input_modules,
    )
    output_modules, output_types = get_output_modules(
        outputs_as_dict=outputs_as_dict,
        device=global_config.be.device,
        computed_out_dimensions=computed_out_dimension,
        feature_dimensions_and_types=feature_dims_and_types,
        fusion_model_type=fusion_config.model_type,
    )
    fusion_to_output_mapping = _match_fusion_outputs_to_output_types(
        output_types=output_types,
        diffusion_targets=diffusion_targets,
    )
    kwargs["output_modules"] = output_modules
    kwargs["fusion_to_output_mapping"] = fusion_to_output_mapping

    input_configs = [i.input_config for i in inputs_as_dict.values()]
    output_configs = [i.output_config for i in outputs_as_dict.values()]
    tensor_broker = get_tensor_broker(
        input_objects=inputs_as_dict,
        output_objects=outputs_as_dict,
        input_modules=input_modules,
        fusion_modules=fusion_modules,
        output_modules=output_modules,
        fusion_to_output_mapping=fusion_to_output_mapping,
        input_configs=input_configs,
        fusion_configs=[fusion_config],
        output_configs=output_configs,
        device=global_config.be.device,
    )
    kwargs["tensor_broker"] = tensor_broker

    return kwargs


def _get_output_types(
    outputs_as_dict: "al_output_objects_as_dict",
) -> dict[str, Literal["tabular", "sequence", "array"]]:
    outputs_to_types_mapping: dict[str, Literal["tabular", "sequence", "array"]] = {}

    for output_name, output_object in outputs_as_dict.items():
        output_type = output_object.output_config.output_info.output_type
        outputs_to_types_mapping[output_name] = output_type

    return outputs_to_types_mapping


def _extract_diffusion_targets(
    outputs_as_dict: "al_output_objects_as_dict",
) -> dict[str, bool]:
    diffusion_targets = {}
    for name, output in outputs_as_dict.items():
        match output:
            case ComputedArrayOutputInfo() | ComputedImageOutputInfo():
                if output.diffusion_config is not None:
                    diffusion_targets[name] = True

    return diffusion_targets


def _get_maybe_computed_out_dims(fusion_modules: al_fusion_modules) -> Optional[int]:
    if "computed" in fusion_modules:
        return fusion_modules["computed"].num_out_features

    return None


def _match_fusion_outputs_to_output_types(
    output_types: dict[str, Literal["tabular", "sequence", "array"]],
    diffusion_targets: dict[str, bool],
) -> dict[str, Literal["computed", "pass-through"]]:
    output_name_to_fusion_output_type: dict[
        str, Literal["computed", "pass-through"]
    ] = {}

    for output_name, output_type in output_types.items():
        match output_type:
            case "tabular" | "survival":
                output_name_to_fusion_output_type[output_name] = "computed"
            case "sequence":
                output_name_to_fusion_output_type[output_name] = "pass-through"
            case "array" | "image":
                if output_name in diffusion_targets:
                    output_name_to_fusion_output_type[output_name] = "pass-through"
                else:
                    output_name_to_fusion_output_type[output_name] = "computed"
            case _:
                raise ValueError(f"Unknown output type '{output_type}'.")

    return output_name_to_fusion_output_type


@dataclass()
class FeatureExtractorInfo:
    input_dimension: Union[
        DataDimensions,
        "OmicsDataDimensions",
        "SequenceDataDimensions",
    ]
    output_dimension: int
    input_type: str
    output_shape: Optional[Tuple[int, ...]] = None
    extras: dict = field(default_factory=dict)


def get_all_feature_extractor_dimensions_and_types(
    inputs_as_dict: al_input_objects_as_dict,
    input_modules: al_input_modules,
) -> Dict[str, FeatureExtractorInfo]:
    input_dimensionality_and_types = {}

    out_feature_per_feature_extractor = _get_feature_extractors_num_output_features(
        input_modules=input_modules
    )
    in_features_per_input = _get_feature_extractors_input_dimensions_per_axis(
        inputs_as_dict=inputs_as_dict, input_modules=input_modules
    )

    output_shapes = _maybe_get_feature_extractor_out_shapes(input_modules=input_modules)

    for input_name, input_object in inputs_as_dict.items():

        cur_output_shape = output_shapes.get(input_name)

        extras = {}
        match input_object:
            case ComputedArrayInputInfo() | ComputedImageInputInfo():
                cur_config = input_object.input_config.model_config
                assert isinstance(
                    cur_config, (schemas.ArrayModelConfig, schemas.ImageModelConfig)
                )
                mic = cur_config.model_init_config
                if isinstance(mic, CNNModelConfig):
                    extras["down_every_n_blocks"] = mic.down_sample_every_n_blocks

        input_dimensionality_and_types[input_name] = FeatureExtractorInfo(
            input_dimension=in_features_per_input[input_name],
            output_dimension=out_feature_per_feature_extractor[input_name],
            input_type=input_object.input_config.input_info.input_type,
            output_shape=cur_output_shape,
            extras=extras,
        )

    return input_dimensionality_and_types


def get_input_modules(
    inputs_as_dict: al_input_objects_as_dict,
    device: str,
) -> al_input_modules:
    input_modules = nn.ModuleDict()

    for input_name, inputs_object in inputs_as_dict.items():
        input_model_config = inputs_object.input_config.model_config

        match inputs_object:
            case ComputedOmicsInputInfo():
                assert isinstance(input_model_config, schemas.OmicsModelConfig)
                cur_omics_model = get_omics_model_from_model_config(
                    model_type=input_model_config.model_type,
                    model_init_config=input_model_config.model_init_config,
                    data_dimensions=inputs_object.data_dimensions,
                )
                input_modules[input_name] = cur_omics_model

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                transformers = inputs_object.labels.label_transformers
                input_type_info = inputs_object.input_config.input_type_info
                assert isinstance(input_type_info, TabularInputDataConfig)
                cat_columns = list(input_type_info.input_cat_columns)
                con_columns = list(input_type_info.input_con_columns)

                unique_tabular_values = get_unique_values_from_transformers(
                    transformers=transformers,
                    keys_to_use=cat_columns,
                )

                assert isinstance(input_model_config, schemas.TabularModelConfig)
                tabular_model = get_tabular_model(
                    model_init_config=input_model_config.model_init_config,
                    cat_columns=cat_columns,
                    con_columns=con_columns,
                    device=device,
                    unique_label_values=unique_tabular_values,
                )
                input_modules[input_name] = tabular_model

            case ComputedSequenceInputInfo() | ComputedBytesInputInfo():
                assert isinstance(input_model_config, schemas.SequenceModelConfig)
                num_tokens = len(inputs_object.vocab)
                sequence_model = get_sequence_model(
                    sequence_model_config=input_model_config,
                    model_registry_lookup=get_sequence_model_class,
                    num_tokens=num_tokens,
                    max_length=inputs_object.computed_max_length,
                    embedding_dim=input_model_config.embedding_dim,
                    device=device,
                )
                input_modules[input_name] = sequence_model

            case ComputedImageInputInfo():
                assert isinstance(input_model_config, schemas.ImageModelConfig)
                image_model = get_image_model(
                    model_config=input_model_config,
                    model_registry_lookup=get_image_model_class,
                    data_dimensions=inputs_object.data_dimensions,
                    device=device,
                )
                input_modules[input_name] = image_model

            case ComputedArrayInputInfo():
                assert isinstance(input_model_config, schemas.ArrayModelConfig)
                array_feature_extractor = get_array_input_feature_extractor(
                    model_type=input_model_config.model_type,
                    model_init_config=input_model_config.model_init_config,
                    data_dimensions=inputs_object.data_dimensions,
                )
                cur_array_model = get_array_model(
                    array_feature_extractor=array_feature_extractor,
                    model_config=input_model_config,
                )
                input_modules[input_name] = cur_array_model

            case _:
                raise ValueError(f"Unrecognized input type for object {inputs_object}")

    return cast(al_input_modules, input_modules)


def get_output_modules(
    outputs_as_dict: "al_output_objects_as_dict",
    device: str,
    fusion_model_type: str,
    computed_out_dimensions: Optional[int] = None,
    feature_dimensions_and_types: Optional[Dict[str, FeatureExtractorInfo]] = None,
) -> Tuple[al_output_modules, Dict[str, Literal["tabular", "sequence", "array"]]]:
    output_modules: al_output_modules = cast(al_output_modules, nn.ModuleDict())
    output_types = {}

    for output_name, output_object in outputs_as_dict.items():
        output_type = output_object.output_config.output_info.output_type
        output_model_config = output_object.output_config.model_config
        output_types[output_name] = output_type

        match output_object:
            case ComputedTabularOutputInfo() | ComputedSurvivalOutputInfo():
                assert computed_out_dimensions is not None
                assert isinstance(
                    output_model_config,
                    (
                        schemas.TabularOutputModuleConfig,
                        schemas.SurvivalOutputTypeConfig,
                    ),
                )
                tabular_output_module = get_tabular_output_module_from_model_config(
                    output_model_config=output_model_config,
                    input_dimension=computed_out_dimensions,
                    num_outputs_per_target=output_object.num_outputs_per_target,
                    device=device,
                )
                output_modules[output_name] = tabular_output_module

            case ComputedSequenceOutputInfo():
                assert isinstance(output_model_config, SequenceOutputModuleConfig)
                feat_dims = feature_dimensions_and_types
                assert feat_dims is not None
                sequence_output_module = get_sequence_output_module_from_model_config(
                    output_object=output_object,
                    feature_dimensionalities_and_types=feat_dims,
                    device=device,
                )
                output_modules[output_name] = sequence_output_module

            case ComputedArrayOutputInfo() | ComputedImageOutputInfo():
                assert isinstance(output_model_config, schemas.ArrayOutputModuleConfig)
                feat_dims = feature_dimensions_and_types

                is_diffusion = output_object.diffusion_config is not None
                setup_func = get_array_or_image_output_module_from_model_config
                match is_diffusion:
                    case True:
                        assert feat_dims is not None
                        output_module = setup_func(
                            output_object=output_object,
                            input_dimension=computed_out_dimensions,
                            fusion_model_type=fusion_model_type,
                            feature_extractor_infos=feat_dims,
                            device=device,
                        )
                    case False:
                        assert computed_out_dimensions is not None
                        output_module = setup_func(
                            output_object=output_object,
                            input_dimension=computed_out_dimensions,
                            fusion_model_type=fusion_model_type,
                            feature_extractor_infos=feat_dims,
                            device=device,
                        )
                    case _:
                        raise ValueError()

                output_modules[output_name] = output_module
            case _:
                raise NotImplementedError(
                    "Only tabular, sequence and array outputs are supported"
                )

    return output_modules, output_types


def _get_feature_extractors_num_output_features(
    input_modules: al_input_modules,
) -> Dict[str, int]:
    fusion_in_dims = {name: i.num_out_features for name, i in input_modules.items()}
    return fusion_in_dims


def _maybe_get_feature_extractor_out_shapes(
    input_modules: al_input_modules,
) -> dict[str, tuple[int, ...]]:
    out_shapes = {
        name: i.output_shape
        for name, i in input_modules.items()
        if hasattr(i, "output_shape")
    }

    return out_shapes


def _get_feature_extractors_input_dimensions_per_axis(
    inputs_as_dict: al_input_objects_as_dict,
    input_modules: al_input_modules,
) -> al_data_dimensions:
    fusion_in_dims: al_data_dimensions = {}

    for name, input_object in inputs_as_dict.items():
        input_model_config = input_object.input_config.model_config

        match input_object:
            case ComputedSequenceInputInfo() | ComputedBytesInputInfo():
                assert isinstance(input_model_config, schemas.SequenceModelConfig)
                fusion_in_dims[name] = SequenceDataDimensions(
                    channels=1,
                    height=input_object.computed_max_length,
                    width=input_model_config.embedding_dim,
                )

            case ComputedImageInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ImageInputDataConfig)
                fusion_in_dims[name] = DataDimensions(
                    channels=input_object.num_channels,
                    height=input_type_info.size[0],
                    width=input_type_info.size[-1],
                )

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                input_model = input_modules[name]
                assert isinstance(input_model, SimpleTabularModel)
                fusion_in_dims[name] = DataDimensions(
                    channels=1,
                    height=1,
                    width=input_model.input_dim,
                )

            case ComputedOmicsInputInfo():
                fusion_in_dims[name] = OmicsDataDimensions(
                    **input_object.data_dimensions.__dict__
                )

            case ComputedArrayInputInfo():
                fusion_in_dims[name] = input_object.data_dimensions

            case _:
                raise ValueError(f"Unrecognized input object {input_object}.")

    return fusion_in_dims



from dataclasses import dataclass
from functools import partial, update_wrapper
from pathlib import Path
from typing import Any, Callable, Optional

import numpy as np
import polars as pl
import torch
from PIL.Image import Image

from eir.data_load.data_preparation_modules.input_preparation_wrappers import (
    numpy_dtype_to_polars_dtype,
    typed_partial_for_hook,
)
from eir.data_load.data_preparation_modules.prepare_array import (
    array_load_wrapper,
    prepare_array_data,
)
from eir.data_load.data_preparation_modules.prepare_image import (
    image_load_wrapper,
    prepare_image_data,
)
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import (
    ComputedImageOutputInfo,
    ImageOutputTypeConfig,
)


def prepare_outputs_disk(
    outputs: dict[str, Any],
    output_objects: "al_output_objects_as_dict",
    test_mode: bool,
) -> dict[str, dict[str, torch.Tensor | int | float]]:
    prepared_outputs: dict[str, dict[str, torch.Tensor | int | float]] = {}

    for output_name, output in outputs.items():
        output_object = output_objects[output_name]
        output_source = output_object.output_config.output_info.output_source
        deeplake_inner_key = output_object.output_config.output_info.output_inner_key

        output_prepared: dict[str, torch.Tensor | int | float]
        match output_object:
            case ComputedArrayOutputInfo():
                data_pointer = output[output_name]
                loaded_array = array_load_wrapper(
                    data_pointer=data_pointer,
                    input_source=output_source,
                    deeplake_inner_key=deeplake_inner_key,
                )
                array_prepared = prepare_array_data(
                    array_data=loaded_array,
                    normalization_stats=output_object.normalization_stats,
                )
                output_prepared = {output_name: array_prepared}

            case ComputedImageOutputInfo():
                output_type_info = output_object.output_config.output_type_info
                assert isinstance(output_type_info, ImageOutputTypeConfig)
                data_pointer = output[output_name]
                loaded_image = image_load_wrapper(
                    data_pointer=data_pointer,
                    input_source=output_source,
                    image_mode=output_type_info.mode,
                    deeplake_inner_key=deeplake_inner_key,
                )
                image_prepared = prepare_image_data(
                    image_input_object=output_object,
                    image_data=loaded_image,
                    test_mode=test_mode,
                )
                output_prepared = {output_name: image_prepared}

            case _:
                output_prepared = output

        prepared_outputs[output_name] = output_prepared

    return prepared_outputs


def prepare_outputs_memory(
    outputs: dict[str, Any],
    output_objects: "al_output_objects_as_dict",
    test_mode: bool,
) -> dict[str, dict[str, torch.Tensor | int | float]]:
    prepared_outputs: dict[str, dict[str, torch.Tensor | int | float]] = {}

    for output_name, output in outputs.items():
        output_object = output_objects[output_name]

        match output_object:
            case ComputedArrayOutputInfo():
                loaded_array = output[output_name]
                array_prepared = prepare_array_data(
                    array_data=loaded_array,
                    normalization_stats=output_object.normalization_stats,
                )
                output_prepared: dict[str, torch.Tensor | int | float] = {
                    output_name: array_prepared
                }

            case ComputedImageOutputInfo():
                loaded_image = output[output_name]
                image_prepared = prepare_image_data(
                    image_input_object=output_object,
                    image_data=loaded_image,
                    test_mode=test_mode,
                )
                output_prepared = {output_name: image_prepared}

            case _:
                output_prepared = output

        prepared_outputs[output_name] = output_prepared

    return prepared_outputs


@dataclass
class HookOutput:
    hook_callable: Callable[..., dict[str, np.ndarray | Image]]
    return_dtype: Optional[pl.DataType]


def get_output_data_loading_hooks(
    outputs: al_output_objects_as_dict,
) -> Optional[dict[str, HookOutput]]:
    """
    Note we use object dtypes for images as they are PIL.Image objects.
    """
    mapping = {}

    for output_name, output_object in outputs.items():
        output_info = output_object.output_config.output_info
        common_kwargs = {
            "input_source": output_info.output_source,
            "deeplake_inner_key": output_info.output_inner_key,
        }

        match output_object:
            case ComputedArrayOutputInfo():
                inner_function = typed_partial_for_hook(
                    array_load_wrapper,
                    **common_kwargs,
                )

                final_callable = typed_partial_for_output_hook(
                    _extract_nested_output_and_call,
                    output_name=output_name,
                    function=inner_function,
                )

                arr_shape = output_object.data_dimensions.original_shape
                polars_dtype = numpy_dtype_to_polars_dtype(np_dtype=output_object.dtype)

                mapping[output_name] = HookOutput(
                    hook_callable=final_callable,
                    return_dtype=pl.Array(inner=polars_dtype, shape=arr_shape),
                )

            case ComputedImageOutputInfo():
                output_type_info = output_object.output_config.output_type_info
                assert isinstance(output_type_info, ImageOutputTypeConfig)
                inner_function = typed_partial_for_hook(
                    image_load_wrapper,
                    **common_kwargs,
                    image_mode=output_type_info.mode,
                )

                final_callable = typed_partial_for_output_hook(
                    _extract_nested_output_and_call,
                    output_name=output_name,
                    function=inner_function,
                )

                mapping[output_name] = HookOutput(
                    hook_callable=final_callable,
                    return_dtype=pl.Object(),
                )

    return mapping


def typed_partial_for_output_hook(
    func: Callable[..., Any], *args: Any, **kwargs: Any
) -> Callable[..., dict[str, np.ndarray | Image]]:
    """
    Just to make mypy happy.
    """
    partial_func = partial(func, *args, **kwargs)
    return update_wrapper(partial_func, func)


def _extract_nested_output_and_call(
    output_dict: dict[str, Path | int],
    output_name: str,
    function: Callable[[Path | int], np.ndarray | Image],
) -> dict[str, np.ndarray | Image]:
    extracted_output_data_pointer = output_dict[output_name]

    loaded_output = function(extracted_output_data_pointer)

    return {output_name: loaded_output}



from dataclasses import dataclass
from functools import partial, update_wrapper
from typing import Any, Callable, Dict, Optional, Type

import numpy as np
import polars as pl
import torch
from PIL.Image import Image
from polars.datatypes import DataTypeClass
from polars.datatypes.classes import NumericType

from eir.data_load.data_preparation_modules.prepare_array import (
    array_load_wrapper,
    prepare_array_data,
)
from eir.data_load.data_preparation_modules.prepare_bytes import (
    bytes_load_wrapper,
    prepare_bytes_data,
)
from eir.data_load.data_preparation_modules.prepare_image import (
    image_load_wrapper,
    prepare_image_data,
)
from eir.data_load.data_preparation_modules.prepare_omics import (
    omics_load_wrapper,
    prepare_one_hot_omics_data,
)
from eir.data_load.data_preparation_modules.prepare_sequence import (
    prepare_sequence_data,
    sequence_load_wrapper,
)
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.schemas import (
    ByteInputDataConfig,
    ImageInputDataConfig,
    OmicsInputDataConfig,
    SequenceInputDataConfig,
    al_input_type_info,
)


def prepare_inputs_disk(
    inputs: dict[str, Any],
    inputs_objects: "al_input_objects_as_dict",
    test_mode: bool,
) -> Dict[str, torch.Tensor]:
    prepared_inputs = {}

    for input_name, data_pointer in inputs.items():
        input_object = inputs_objects[input_name]
        input_source = input_object.input_config.input_info.input_source
        input_type_info = input_object.input_config.input_type_info
        deeplake_inner_key = input_object.input_config.input_info.input_inner_key

        drop_rate = _get_modality_drop_rate(input_type_info=input_type_info)
        should_skip = _should_skip_modality(
            modality_dropout_rate=drop_rate,
            test_mode=test_mode,
        )

        if should_skip:
            continue

        match input_object:
            case ComputedOmicsInputInfo():
                array_raw = omics_load_wrapper(
                    input_source=input_source,
                    data_pointer=data_pointer,
                    deeplake_inner_key=deeplake_inner_key,
                    subset_indices=input_object.subset_indices,
                )

                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, OmicsInputDataConfig)
                input_prepared = prepare_one_hot_omics_data(
                    genotype_array=array_raw,
                    na_augment_alpha=input_type_info.na_augment_alpha,
                    na_augment_beta=input_type_info.na_augment_beta,
                    shuffle_augment_alpha=input_type_info.shuffle_augment_alpha,
                    shuffle_augment_beta=input_type_info.shuffle_augment_beta,
                    test_mode=test_mode,
                )

            case ComputedSequenceInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)
                sequence_tokenized = sequence_load_wrapper(
                    data_pointer=data_pointer,
                    input_source=input_source,
                    deeplake_inner_key=deeplake_inner_key,
                    split_on=input_type_info.split_on,
                    encode_func=input_object.encode_func,
                )
                input_prepared = prepare_sequence_data(
                    sequence_input_object=input_object,
                    cur_file_content_tokenized=sequence_tokenized,
                    test_mode=test_mode,
                )

            case ComputedBytesInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ByteInputDataConfig)
                bytes_data = bytes_load_wrapper(
                    data_pointer=data_pointer,
                    dtype=input_type_info.byte_encoding,
                    input_source=input_source,
                    deeplake_inner_key=deeplake_inner_key,
                )
                input_prepared = prepare_bytes_data(
                    bytes_input_object=input_object,
                    bytes_data=bytes_data,
                    test_mode=test_mode,
                )

            case ComputedImageInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ImageInputDataConfig)
                image_data = image_load_wrapper(
                    input_source=input_source,
                    data_pointer=data_pointer,
                    image_mode=input_type_info.mode,
                    deeplake_inner_key=deeplake_inner_key,
                )
                input_prepared = prepare_image_data(
                    image_input_object=input_object,
                    image_data=image_data,
                    test_mode=test_mode,
                )

            case ComputedArrayInputInfo():
                array_data = array_load_wrapper(
                    input_source=input_source,
                    data_pointer=data_pointer,
                    deeplake_inner_key=deeplake_inner_key,
                )
                input_prepared = prepare_array_data(
                    array_data=array_data,
                    normalization_stats=input_object.normalization_stats,
                )

            case _:
                input_prepared = inputs[input_name]

        prepared_inputs[input_name] = input_prepared

    return prepared_inputs


def _get_modality_drop_rate(input_type_info: al_input_type_info) -> float:
    drop_rate = getattr(input_type_info, "modality_dropout_rate", 0.0)
    return drop_rate


def _should_skip_modality(modality_dropout_rate: float, test_mode: bool) -> bool:
    if test_mode or modality_dropout_rate == 0.0:
        return False

    should_skip = torch.rand(1) < modality_dropout_rate
    return bool(should_skip)


def prepare_inputs_memory(
    inputs: Dict[str, Any],
    inputs_objects: "al_input_objects_as_dict",
    test_mode: bool,
) -> Dict[str, torch.Tensor]:
    prepared_inputs = {}

    for name, data in inputs.items():
        input_object = inputs_objects[name]
        input_type_info = input_object.input_config.input_type_info

        drop_rate = _get_modality_drop_rate(input_type_info=input_type_info)
        should_skip = _should_skip_modality(
            modality_dropout_rate=drop_rate,
            test_mode=test_mode,
        )

        if should_skip:
            continue

        match input_object:
            case ComputedOmicsInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, OmicsInputDataConfig)
                input_prepared = prepare_one_hot_omics_data(
                    genotype_array=data,
                    na_augment_alpha=input_type_info.na_augment_alpha,
                    na_augment_beta=input_type_info.na_augment_beta,
                    shuffle_augment_alpha=input_type_info.shuffle_augment_alpha,
                    shuffle_augment_beta=input_type_info.shuffle_augment_beta,
                    test_mode=test_mode,
                )

            case ComputedSequenceInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)
                input_prepared = prepare_sequence_data(
                    sequence_input_object=input_object,
                    cur_file_content_tokenized=data,
                    test_mode=test_mode,
                )

            case ComputedBytesInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ByteInputDataConfig)
                input_prepared = prepare_bytes_data(
                    bytes_input_object=input_object,
                    bytes_data=data,
                    test_mode=test_mode,
                )

            case ComputedImageInputInfo():
                input_prepared = prepare_image_data(
                    image_input_object=input_object,
                    image_data=data,
                    test_mode=test_mode,
                )

            case ComputedArrayInputInfo():
                input_prepared = prepare_array_data(
                    array_data=data,
                    normalization_stats=input_object.normalization_stats,
                )

            case _:
                input_prepared = inputs[name]

        prepared_inputs[name] = input_prepared

    return prepared_inputs


def typed_partial_for_hook(
    func: Callable[..., Any], *args: Any, **kwargs: Any
) -> Callable[..., np.ndarray | Image]:
    """
    Just to make mypy happy.
    """
    partial_func = partial(func, *args, **kwargs)
    return update_wrapper(partial_func, func)


@dataclass
class InputHookOutput:
    hook_callable: Callable[..., np.ndarray | Image | str]
    return_dtype: Optional[pl.DataType]


def get_input_data_loading_hooks(
    inputs: al_input_objects_as_dict,
) -> Optional[dict[str, InputHookOutput]]:
    """
    Creates data loading hooks for input objects with appropriate polars dtypes.
    Note we use object dtypes for images as they are PIL.Image objects.
    """
    mapping = {}

    for input_name, input_object in inputs.items():
        common_kwargs = {
            "input_source": input_object.input_config.input_info.input_source,
            "deeplake_inner_key": input_object.input_config.input_info.input_inner_key,
        }

        match input_object:
            case ComputedOmicsInputInfo():
                inner_function = typed_partial_for_hook(
                    omics_load_wrapper,
                    **common_kwargs,
                    subset_indices=input_object.subset_indices,
                )

                # skip the batch dimension here for polars later as loading
                # the array will return a 2D array
                arr_shape = input_object.data_dimensions.full_shape()[1:]
                mapping[input_name] = InputHookOutput(
                    hook_callable=inner_function,
                    return_dtype=pl.Array(inner=pl.Boolean, shape=arr_shape),
                )

            case ComputedImageInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ImageInputDataConfig)

                inner_function = typed_partial_for_hook(
                    image_load_wrapper,
                    **common_kwargs,
                    image_mode=input_type_info.mode,
                )

                mapping[input_name] = InputHookOutput(
                    hook_callable=inner_function,
                    return_dtype=pl.Object(),
                )

            case ComputedSequenceInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)

                inner_function = typed_partial_for_hook(
                    sequence_load_wrapper,
                    **common_kwargs,
                    split_on=input_type_info.split_on,
                    encode_func=input_object.encode_func,
                )

                mapping[input_name] = InputHookOutput(
                    hook_callable=inner_function,
                    return_dtype=pl.List(pl.Int64),
                )

            case ComputedBytesInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ByteInputDataConfig)

                inner_function = typed_partial_for_hook(
                    bytes_load_wrapper,
                    **common_kwargs,
                    dtype=input_type_info.byte_encoding,
                )

                pl_dtype = _get_polars_dtype_for_bytes(
                    byte_encoding=input_type_info.byte_encoding
                )
                max_length = input_object.computed_max_length
                mapping[input_name] = InputHookOutput(
                    hook_callable=inner_function,
                    return_dtype=pl.Array(inner=pl_dtype, shape=(max_length,)),
                )

            case ComputedArrayInputInfo():
                inner_function = typed_partial_for_hook(
                    array_load_wrapper,
                    **common_kwargs,
                )

                polars_dtype = numpy_dtype_to_polars_dtype(np_dtype=input_object.dtype)
                loaded_shape = input_object.data_dimensions.original_shape

                mapping[input_name] = InputHookOutput(
                    hook_callable=inner_function,
                    return_dtype=pl.Array(inner=polars_dtype, shape=loaded_shape),
                )

    return mapping


def _get_polars_dtype_for_bytes(
    byte_encoding: str,
) -> Type[NumericType] | Type[pl.Object]:
    encoding_dtype_map = {
        "uint8": pl.UInt8,
        "int8": pl.Int8,
        "float32": pl.Float32,
    }
    return encoding_dtype_map.get(byte_encoding, pl.Object)


def numpy_dtype_to_polars_dtype(np_dtype: np.dtype) -> DataTypeClass:

    dtype_str = str(np_dtype)

    dtype_map = {
        "float64": pl.Float64,
        "float32": pl.Float32,
        "int64": pl.Int64,
        "int32": pl.Int32,
        "int16": pl.Int16,
        "int8": pl.Int8,
        "uint64": pl.UInt64,
        "uint32": pl.UInt32,
        "uint16": pl.UInt16,
        "uint8": pl.UInt8,
        "bool": pl.Boolean,
        "object": pl.Object,
        "string": pl.String,
        "datetime64[ns]": pl.Datetime,
    }

    return dtype_map.get(dtype_str, pl.Float32)



import json
from copy import deepcopy
from pathlib import Path

from eir.experiment_io.input_object_io_modules.input_io_utils import (
    load_input_config_from_yaml,
)
from eir.experiment_io.io_utils import load_dataclass
from eir.setup import schemas
from eir.setup.input_setup_modules.setup_image import (
    ComputedImageInputInfo,
    ImageNormalizationStats,
    set_up_computed_image_input_object,
)


def load_image_input_object(
    serialized_input_folder: Path,
) -> ComputedImageInputInfo:

    config_path = serialized_input_folder / "input_config.yaml"
    normalization_stats_path = serialized_input_folder / "normalization_stats.json"
    num_channels_path = serialized_input_folder / "num_channels.json"

    input_config = load_input_config_from_yaml(input_config_path=config_path)
    input_type_info_modified = deepcopy(input_config.input_type_info)
    assert isinstance(input_type_info_modified, schemas.ImageInputDataConfig)

    normalization_stats = load_dataclass(
        cls=ImageNormalizationStats, file_path=normalization_stats_path
    )

    image_mode = input_type_info_modified.mode
    if not image_mode:
        num_channels = json.loads(num_channels_path.read_text())["num_channels"]
        input_type_info_modified.num_channels = num_channels
    else:
        input_type_info_modified.num_channels = None

    input_config_modified = deepcopy(input_config)
    input_config_modified.input_type_info = input_type_info_modified

    loaded_object = set_up_computed_image_input_object(
        input_config=input_config_modified,
        normalization_stats=normalization_stats,
    )

    return loaded_object



from collections import OrderedDict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, Sequence

import torch
from torch import nn

from eir.models.fusion.fusion_default import al_features, default_fuse_features
from eir.models.layers.mlp_layers import MLPResidualBlock
from eir.models.models_utils import (
    calculate_module_dict_outputs,
    construct_multi_branches,
    create_multi_task_blocks_with_first_adaptor_block,
    initialize_modules_from_spec,
)

if TYPE_CHECKING:
    pass


@dataclass
class MGMoEModelConfig:
    """
    :param layers:
        A sequence of two int values controlling the number of residual MLP blocks in
        the network. The first item (i.e. ``layers[0]``) refers to the number of blocks
        in the expert branches. The second item (i.e. ``layers[1]``) refers to the
        number of blocks in the predictor branches.

    :param fc_task_dim:
       Number of hidden nodes in all residual blocks (both expert and predictor) of
       the network.

    :param mg_num_experts:
        Number of multi gate experts to use.

    :param rb_do:
        Dropout in all MLP residual blocks (both expert and predictor).

    :param fc_do:
        Dropout before the last FC layer.

    :param stochastic_depth_p:
        Probability of dropping input.
    """

    layers: Sequence[int] = field(default_factory=lambda: [1, 1])
    fc_task_dim: int = 64

    mg_num_experts: int = 8

    rb_do: float = 0.00
    fc_do: float = 0.00

    stochastic_depth_p: float = 0.00


class MGMoEModel(nn.Module):
    def __init__(
        self,
        model_config: MGMoEModelConfig,
        fusion_in_dim: int,
        fusion_callable: al_features = default_fuse_features,
        **kwargs,
    ):
        super().__init__()

        self.model_config = model_config
        self.fusion_in_dim = fusion_in_dim
        self.fusion_callable = fusion_callable

        self.num_experts = self.model_config.mg_num_experts

        gate_spec = self.get_gate_spec(
            in_features=self.fusion_in_dim, out_features=self.num_experts
        )

        expert_names = tuple(f"expert_{i}" for i in range(self.num_experts))
        layer_kwargs = {
            "in_features": self.model_config.fc_task_dim,
            "out_features": self.model_config.fc_task_dim,
            "dropout_p": self.model_config.rb_do,
            "stochastic_depth_p": self.model_config.stochastic_depth_p,
            "full_preactivation": False,
        }
        self.expert_branches = create_multi_task_blocks_with_first_adaptor_block(
            num_blocks=self.model_config.layers[0],
            branch_names=expert_names,
            block_constructor=MLPResidualBlock,
            block_constructor_kwargs=layer_kwargs,
            first_layer_kwargs_overload={
                "full_preactivation": True,
                "in_features": fusion_in_dim,
            },
        )

        self.gates = construct_multi_branches(
            branch_names=expert_names,
            branch_factory=initialize_modules_from_spec,
            branch_factory_kwargs={"spec": gate_spec},
        )

        self._init_weights()

    @staticmethod
    def get_gate_spec(in_features: int, out_features: int):
        spec = OrderedDict(
            {
                "gate_fc": (
                    nn.Linear,
                    {
                        "in_features": in_features,
                        "out_features": out_features,
                        "bias": True,
                    },
                ),
                "gate_attention": (nn.Softmax, {"dim": 1}),
            }
        )

        return spec

    def _init_weights(self):
        pass

    @property
    def num_out_features(self) -> int:
        return self.model_config.fc_task_dim * len(self.expert_branches)

    def forward(self, inputs: Dict[str, torch.Tensor]) -> torch.Tensor:
        fused_features = self.fusion_callable(inputs)

        gate_attentions = calculate_module_dict_outputs(
            input_=fused_features, module_dict=self.gates
        )

        expert_outputs = calculate_module_dict_outputs(
            input_=fused_features, module_dict=self.expert_branches
        )

        final_out = {}
        stacked_expert_outputs = torch.stack(list(expert_outputs.values()), dim=2)
        for expert_name, attention in gate_attentions.items():
            weighted_expert_outputs = attention.unsqueeze(1) * stacked_expert_outputs
            weighted_expert_sum = weighted_expert_outputs.sum(dim=2)

            final_out[expert_name] = weighted_expert_sum

        return self.fusion_callable(final_out)



from pathlib import Path
from typing import Optional, Sequence, Union

import numpy as np
import torch

from eir.data_load.data_preparation_modules.common import (
    _load_deeplake_sample,
    process_tensor_to_length,
)
from eir.data_load.data_source_modules import deeplake_ops
from eir.setup.input_setup_modules.setup_sequence import (
    ComputedSequenceInputInfo,
    al_encode_funcs,
    get_sequence_split_function,
)
from eir.setup.schemas import SequenceInputDataConfig


def sequence_load_wrapper(
    data_pointer: Union[Path, int, np.ndarray],
    input_source: str,
    split_on: Optional[str],
    encode_func: al_encode_funcs,
    deeplake_inner_key: Optional[str] = None,
) -> np.ndarray:
    """
    In the case of .csv input sources, we have already loaded and tokenized the data.
    """

    split_func = get_sequence_split_function(split_on=split_on)
    if deeplake_ops.is_deeplake_dataset(data_source=input_source):
        assert deeplake_inner_key is not None
        assert isinstance(data_pointer, int)
        content = _load_deeplake_sample(
            data_pointer=data_pointer,
            input_source=input_source,
            inner_key=deeplake_inner_key,
        )
        assert isinstance(content, str)
    elif input_source.endswith(".csv"):
        assert isinstance(data_pointer, np.ndarray)
        return data_pointer
    else:
        assert isinstance(data_pointer, (str, Path))
        content = load_sequence_from_disk(sequence_file_path=data_pointer)

    file_content_split = split_func(content)
    file_content_encoded = encode_func(file_content_split)
    sequence_tokenized = np.array(file_content_encoded)

    return sequence_tokenized


def load_sequence_from_disk(sequence_file_path: Path | str) -> str:
    with open(sequence_file_path, "r") as infile:
        return infile.read().strip()


def prepare_sequence_data(
    sequence_input_object: "ComputedSequenceInputInfo",
    cur_file_content_tokenized: np.ndarray,
    test_mode: bool,
) -> torch.Tensor:
    """
    We use clone here to copy the original data, vs. using from_numpy
    which shares memory, causing us to modify the original data.
    """

    sio = sequence_input_object
    input_type_info = sio.input_config.input_type_info
    assert isinstance(input_type_info, SequenceInputDataConfig)

    cur_arr = cur_file_content_tokenized.copy()
    cur_tokens_as_tensor = torch.LongTensor(cur_arr).detach().clone()

    sampling_strategy = input_type_info.sampling_strategy_if_longer
    if test_mode:
        sampling_strategy = "from_start"

    padding_token = getattr(sio.tokenizer, "pad_token", "<pad>")
    padding_token_parsed = parse_padding_token_encode_func_input(
        split_on=input_type_info.split_on, padding_token=padding_token
    )
    padding_value = sio.encode_func(padding_token_parsed)[0]
    cur_tokens_padded = process_tensor_to_length(
        tensor=cur_tokens_as_tensor,
        max_length=sio.computed_max_length,
        sampling_strategy_if_longer=sampling_strategy,
        padding_value=padding_value,
    )

    return cur_tokens_padded


def parse_padding_token_encode_func_input(
    split_on: Optional[str], padding_token: str
) -> Sequence[str] | str:
    parsed_token: Sequence[str] | str

    if split_on is None:
        parsed_token = padding_token
    else:
        parsed_token = [padding_token]

    return parsed_token



from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Dict, Literal, Optional, Type, Union

import torch
from torch import nn

from eir.models.input.array.models_cnn import CNNModel, CNNModelConfig
from eir.models.input.array.models_locally_connected import (
    FlattenFunc,
    LCLModel,
    LCLModelConfig,
)
from eir.models.input.array.models_transformers import (
    ArrayTransformer,
    ArrayTransformerConfig,
)

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions

al_array_model_types = Literal["cnn", "lcl", "transformer"]

al_array_model_classes = Type[CNNModel] | Type[LCLModel] | Type[ArrayTransformer]
al_array_models = CNNModel | LCLModel | ArrayTransformer

al_array_model_config_classes = (
    Type[CNNModelConfig] | Type[LCLModelConfig] | Type[ArrayTransformerConfig]
)
al_array_model_configs = CNNModelConfig | LCLModelConfig | ArrayTransformerConfig

al_pre_normalization = Optional[Literal["instancenorm", "layernorm"]]

al_array_model_init_kwargs = dict[
    str,
    Union[
        "DataDimensions",
        CNNModelConfig,
        LCLModelConfig,
        ArrayTransformerConfig,
        FlattenFunc,
    ],
]


def get_array_model_mapping() -> Dict[str, al_array_model_classes]:
    mapping = {
        "cnn": CNNModel,
        "lcl": LCLModel,
        "transformer": ArrayTransformer,
    }

    return mapping


def get_array_model_class(model_type: al_array_model_types) -> al_array_model_classes:
    mapping = get_array_model_mapping()
    return mapping[model_type]


def get_array_config_dataclass_mapping() -> Dict[str, al_array_model_config_classes]:
    mapping = {
        "cnn": CNNModelConfig,
        "lcl": LCLModelConfig,
        "transformer": ArrayTransformerConfig,
    }

    return mapping


def get_model_config_dataclass(model_type: str) -> al_array_model_config_classes:
    mapping = get_array_config_dataclass_mapping()
    return mapping[model_type]


def get_array_model_init_kwargs(
    model_type: al_array_model_types,
    model_config: al_array_model_configs,
    data_dimensions: "DataDimensions",
) -> al_array_model_init_kwargs:
    kwargs: al_array_model_init_kwargs = {}

    model_config_dataclass = get_model_config_dataclass(model_type=model_type)
    model_config_dataclass_instance = model_config_dataclass(**model_config.__dict__)

    kwargs["model_config"] = model_config_dataclass_instance
    kwargs["data_dimensions"] = data_dimensions

    match model_type:
        case "lcl":
            assert isinstance(model_config, LCLModelConfig)

            if model_config.patch_size is not None:
                assert isinstance(model_config.patch_size, (tuple, list))
                assert len(model_config.patch_size) == 3, model_config.patch_size
                kwargs["flatten_fn"] = partial(
                    patchify_and_flatten,
                    size=model_config.patch_size,
                )
            else:
                kwargs["flatten_fn"] = partial(torch.flatten, start_dim=1)

        case "transformer":
            assert isinstance(model_config, ArrayTransformerConfig)
            kwargs["flatten_fn"] = partial(
                patchify,
                size=model_config.patch_size,
                stride=model_config.patch_size,
            )

    return kwargs


def check_patch_and_input_size_compatibility(
    patch_size: Union[tuple[int, int, int], list[int]],
    data_dimensions: "DataDimensions",
) -> None:
    assert isinstance(patch_size, (tuple, list))
    assert len(patch_size) == 3, patch_size

    channels, height, width = patch_size

    if (
        data_dimensions.channels % channels != 0
        or data_dimensions.height % height != 0
        or data_dimensions.width % width != 0
    ):
        mismatch_details = (
            f"Data dimensions {data_dimensions.full_shape()} "
            f"cannot be evenly divided into patches of size {patch_size}. "
            f"Mismatch in channels: {data_dimensions.channels % channels}, "
            f"height: {data_dimensions.height % height}, "
            f"width: {data_dimensions.width % width}."
        )
        raise ValueError(mismatch_details)


def patchify_and_flatten(
    x: torch.Tensor,
    size: tuple[int, int, int],
) -> torch.Tensor:
    stride = size
    patches = patchify(x=x, size=size, stride=stride)
    flattened = flatten_patches(patches=patches)
    return flattened


def patchify(
    x: torch.Tensor, size: tuple[int, int, int], stride: tuple[int, int, int]
) -> torch.Tensor:
    """

    size: (C, H, W)

    Input shape: [256, 3, 64, 64]

    Batch size: batch_size
    Channels: C
    Vertical patches: height / H
    Horizontal patches: width / W
    Patch channels: C
    Patch height: H
    Patch width: W

    After unfolding: [batch_size, C, height / H, width / W, C, H, W]

    After permuting: [batch_size, height / H, width / W, C, C, H, W]
    """
    patches = (
        x.unfold(1, size[0], stride[0])
        .unfold(2, size[1], stride[1])
        .unfold(3, size[2], stride[2])
    )
    patches = patches.permute(0, 2, 3, 4, 1, 5, 6)
    return patches


def flatten_patches(patches: torch.Tensor) -> torch.Tensor:
    reshaped_patches = patches.reshape(patches.size(0), -1)
    return reshaped_patches


@dataclass
class ArrayModelConfig:
    """
    :param model_type:
         Which type of image model to use.

    :param model_init_config:
          Configuration used to initialise model.
    """

    model_type: al_array_model_types
    model_init_config: al_array_model_configs
    pre_normalization: al_pre_normalization = None


class ArrayWrapperModel(nn.Module):
    def __init__(
        self,
        feature_extractor: al_array_models,
        normalization: al_pre_normalization,
    ):
        super().__init__()
        self.feature_extractor = feature_extractor
        self.data_dimensions = feature_extractor.data_dimensions
        self.pre_normalization = get_pre_normalization_layer(
            normalization=normalization, data_dimensions=self.data_dimensions
        )
        self.output_shape = self.feature_extractor.output_shape

    @property
    def num_out_features(self):
        return self.feature_extractor.num_out_features

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.feature_extractor.l1_penalized_weights

    def forward(self, x):
        out = self.pre_normalization(x)
        out = self.feature_extractor(out)
        return out


def get_pre_normalization_layer(
    normalization: al_pre_normalization,
    data_dimensions: "DataDimensions",
) -> Union[nn.InstanceNorm2d, nn.RMSNorm, nn.Identity]:
    channels = data_dimensions.channels
    height = data_dimensions.height
    width = data_dimensions.width

    match normalization:
        case "instancenorm":
            return nn.InstanceNorm2d(
                num_features=channels, affine=True, track_running_stats=True
            )
        case "layernorm":
            return nn.RMSNorm(normalized_shape=[channels, height, width])
        case None:
            return nn.Identity()



from collections import OrderedDict
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import torch
from torch import nn
from transformers import PreTrainedModel

from eir.models.input.sequence.transformer_models import get_hf_transformer_forward
from eir.models.layers.mlp_layers import MLPResidualBlock
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.output_setup_modules.tabular_output_setup import (
        al_num_outputs_per_target,
    )

logger = get_logger(name=__name__, tqdm_compatible=True)


def merge_module_dicts(module_dicts: Tuple[nn.ModuleDict, ...]) -> nn.ModuleDict:
    def _check_inputs():
        assert all(i.keys() == module_dicts[0].keys() for i in module_dicts)

    _check_inputs()

    new_module_dicts = deepcopy(module_dicts)
    final_module_dict = nn.ModuleDict()

    keys = new_module_dicts[0].keys()
    for key in keys:
        final_module_dict[key] = nn.Sequential()

        for index, module_dict in enumerate(new_module_dicts):
            cur_module = module_dict[key]
            final_module_dict[key].add_module(str(index), cur_module)

    return final_module_dict


def construct_blocks(
    num_blocks: int,
    block_constructor: Callable,
    block_kwargs: Dict,
) -> nn.Sequential:
    blocks = []
    for i in range(num_blocks):
        cur_block = block_constructor(**block_kwargs)
        blocks.append(cur_block)
    return nn.Sequential(*blocks)


def create_multi_task_blocks_with_first_adaptor_block(
    num_blocks: int,
    branch_names: Sequence[str],
    block_constructor: Callable,
    block_constructor_kwargs: dict,
    first_layer_kwargs_overload: dict,
) -> nn.ModuleDict:
    adaptor_block = construct_multi_branches(
        branch_names=branch_names,
        branch_factory=construct_blocks,
        branch_factory_kwargs={
            "num_blocks": 1,
            "block_constructor": block_constructor,
            "block_kwargs": {**block_constructor_kwargs, **first_layer_kwargs_overload},
        },
    )

    if num_blocks == 1:
        return merge_module_dicts((adaptor_block,))

    blocks = construct_multi_branches(
        branch_names=branch_names,
        branch_factory=construct_blocks,
        branch_factory_kwargs={
            "num_blocks": num_blocks - 1,
            "block_constructor": block_constructor,
            "block_kwargs": {**block_constructor_kwargs},
        },
    )

    merged_blocks = merge_module_dicts(module_dicts=(adaptor_block, blocks))

    return merged_blocks


def assert_module_dict_uniqueness(module_dict: nn.ModuleDict):
    """
    We have this function as a safeguard to help us catch if we are reusing modules
    when they should not be (i.e. if splitting into multiple branches with same layers,
    one could accidentally reuse the instantiated nn.Modules across branches).
    """
    branch_ids = [id(sequential_branch) for sequential_branch in module_dict.values()]
    assert len(branch_ids) == len(set(branch_ids))

    module_ids = []
    for sequential_branch in module_dict.values():
        module_ids += [id(module) for module in sequential_branch.modules()]

    num_total_modules = len(module_ids)
    num_unique_modules = len(set(module_ids))
    assert num_unique_modules == num_total_modules


def construct_multi_branches(
    branch_names: Iterable[str],
    branch_factory: Callable[..., nn.Sequential],
    branch_factory_kwargs: dict[str, Any],
    extra_hooks: Sequence[Callable] = (),
) -> nn.ModuleDict:
    branched_module_dict = nn.ModuleDict()
    for name in branch_names:
        cur_branch = branch_factory(**branch_factory_kwargs)
        assert callable(cur_branch)
        branched_module_dict[name] = cur_branch

    for hook in extra_hooks:
        branched_module_dict = hook(branched_module_dict)

    assert_module_dict_uniqueness(module_dict=branched_module_dict)
    return branched_module_dict


def get_final_layer(
    in_features: int,
    num_outputs_per_target: "al_num_outputs_per_target",
    layer_type: Union[Literal["linear"], Literal["mlp_residual"]] = "linear",
    layer_type_specific_kwargs: Optional[Dict] = None,
) -> nn.ModuleDict:
    final_module_dict = nn.ModuleDict()

    if layer_type_specific_kwargs is None:
        layer_type_specific_kwargs = {}

    for task, num_outputs in num_outputs_per_target.items():
        if layer_type == "mlp_residual":
            spec_init_kwargs = _get_mlp_residual_final_spec(
                in_features=in_features,
                num_outputs=num_outputs,
                dropout_p=layer_type_specific_kwargs.get("dropout_p", 0.0),
                stochastic_depth_p=layer_type_specific_kwargs.get(
                    "stochastic_depth_p", 0.0
                ),
            )
        else:
            spec_init_kwargs = _get_mlp_basic_final_spec(
                in_features=in_features,
                num_outputs=num_outputs,
                bias=layer_type_specific_kwargs.get("bias", True),
            )

        cur_spec = OrderedDict({"final": spec_init_kwargs})
        cur_module = initialize_modules_from_spec(spec=cur_spec)
        final_module_dict[task] = cur_module

    return final_module_dict


def _get_mlp_residual_final_spec(
    in_features: int,
    num_outputs: int,
    dropout_p: float,
    stochastic_depth_p: float,
) -> Tuple[Type[nn.Module], Dict[str, Any]]:
    spec = (
        MLPResidualBlock,
        {
            "in_features": in_features,
            "out_features": num_outputs,
            "dropout_p": dropout_p,
            "stochastic_depth_p": stochastic_depth_p,
        },
    )

    return spec


def _get_mlp_basic_final_spec(
    in_features: int,
    num_outputs: int,
    bias: bool = True,
) -> Tuple[Type[nn.Module], Dict[str, Any]]:
    spec = (
        nn.Linear,
        {
            "in_features": in_features,
            "out_features": num_outputs,
            "bias": bias,
        },
    )

    return spec


def initialize_modules_from_spec(
    spec: "OrderedDict[str, Tuple[Type[nn.Module], Dict]]",
) -> nn.Sequential:
    module_dict = OrderedDict()
    for name, recipe in spec.items():
        module_class = recipe[0]
        module_args = recipe[1]

        module = initialize_module(module=module_class, module_args=module_args)

        module_dict[name] = module

    return nn.Sequential(module_dict)


def initialize_module(module: Type[nn.Module], module_args: Dict) -> nn.Module:
    return module(**module_args)


def calculate_module_dict_outputs(
    input_: torch.Tensor, module_dict: nn.ModuleDict
) -> "OrderedDict[str, torch.Tensor]":
    out = OrderedDict()
    for name, module in module_dict.items():
        out[name] = module(input_)

    return out


def get_output_dimensions_for_input(
    module: Union[PreTrainedModel, nn.Module],
    input_shape: Tuple[int, ...],
    pool: Union[Literal["max"], Literal["avg"], None],
    hf_model: bool = False,
) -> torch.LongTensor:
    cpu_device = torch.device("cpu")
    module_copy = deepcopy(module).to(device=cpu_device)

    with torch.no_grad():
        test_input = torch.randn(*input_shape)

        if hf_model:
            hf_forward = get_hf_transformer_forward(
                feature_extractor_=module_copy,
                input_length=input_shape[1],
                embedding_dim=input_shape[2],
                device="cpu",
                pool=pool,
            )
            test_output = hf_forward(input=test_input, feature_extractor=module_copy)
        else:
            test_output = module_copy(test_input)

        output_shape = test_output.shape
        logger.info(
            "Inferred output shape of %s from model %s.",
            output_shape,
            type(module).__name__,
        )
        return output_shape



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict

import torch
from torch import nn

from eir.models.layers.mlp_layers import MLPResidualBlock, ResidualMLPConfig
from eir.models.models_utils import create_multi_task_blocks_with_first_adaptor_block

if TYPE_CHECKING:
    from eir.setup.output_setup_modules.tabular_output_setup import (
        al_num_outputs_per_target,
    )


@dataclass
class SharedResidualMLPOutputModuleConfig(ResidualMLPConfig):
    """
    :param layers:
        Number of residual MLP residual blocks to use in the output module.

    :param fc_task_dim:
        Number of hidden nodes in each MLP residual block.

    :param rb_do:
        Dropout in each MLP residual block.

    :param fc_do:
        Dropout before final layer.

    :param stochastic_depth_p:
        Stochastic depth probability (probability of dropping input)
        for each residual block.
    """


class SharedResidualMLPOutputModule(nn.Module):
    def __init__(
        self,
        model_config: SharedResidualMLPOutputModuleConfig,
        input_dimension: int,
        num_outputs_per_target: "al_num_outputs_per_target",
    ):
        super().__init__()

        self.model_config = model_config
        self.input_dimension = input_dimension
        self.num_outputs_per_target = num_outputs_per_target

        self.total_outputs = sum(self.num_outputs_per_target.values())
        sorted_targets = sorted(num_outputs_per_target.items())
        target_names, target_sizes = zip(*sorted_targets)

        self.target_names = target_names
        self.target_sizes = list(target_sizes)

        task_resblocks_kwargs: dict[str, float | int | bool] = {
            "in_features": self.model_config.fc_task_dim,
            "out_features": self.model_config.fc_task_dim,
            "dropout_p": self.model_config.rb_do,
            "stochastic_depth_p": self.model_config.stochastic_depth_p,
            "full_preactivation": False,
        }

        shared_branch_dict = create_multi_task_blocks_with_first_adaptor_block(
            num_blocks=self.model_config.layers[0],
            branch_names=("shared",),
            block_constructor=MLPResidualBlock,
            block_constructor_kwargs=task_resblocks_kwargs,
            first_layer_kwargs_overload={"in_features": self.input_dimension},
        )

        assert len(shared_branch_dict) == 1
        shared_branch_module = shared_branch_dict["shared"]

        final_block = MLPResidualBlock(
            in_features=self.model_config.fc_task_dim,
            out_features=self.total_outputs,
            dropout_p=self.model_config.rb_do,
            stochastic_depth_p=self.model_config.stochastic_depth_p,
            full_preactivation=False,
        )

        self.shared_branch = nn.Sequential(shared_branch_module, final_block)

    def forward(self, inputs: torch.Tensor) -> Dict[str, torch.Tensor]:

        shared_out_tensor = self.shared_branch(inputs)

        split_outputs = torch.split(shared_out_tensor, self.target_sizes, dim=1)

        return {
            target: output for target, output in zip(self.target_names, split_outputs)
        }



from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

import torch
from sympy import Symbol
from sympy.solvers import solve
from torch import nn

from eir.models.layers.cnn_layers import (
    CNNResidualBlock,
    ConvAttentionBlock,
    DownSamplingResidualBlock,
    FirstCNNBlock,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.input_setup_modules.common import DataDimensions

logger = get_logger(__name__)


@dataclass
class CNNModelConfig:
    """
    :param layers:
        A list that controls the number of layers and channels in the model.
        Each element in the list represents a layer group with a specified number of
        layers and channels. Specifically,

        - The first element in the list refers to the number of layers with the
          number of channels exactly as specified by the ``channel_exp_base`` parameter.

        - The subsequent elements in the list correspond to an increased number
          of channels, doubling with each step. For instance, if ``channel_exp_base=3``
          (i.e., ``2**3=8`` channels), and the ``layers`` list is ``[5, 3, 2]``,
          the model would be constructed as follows,

            - First case: 5 layers with 8 channels
            - Second case: 3 layers with 16 channels (doubling from the previous case)
            - Third case: 2 layers with 32 channels (doubling from the previous case)

        - The model currently supports a maximum of 4 elements in the list.

        - If set to ``None``, the model will automatically set up the number of
          layer groups until a certain width and height (``stride * 8`` for both)
          are met. In this automatic setup, channels will be increased as the input
          gets propagated through the network, while the width/height get reduced
          due to stride.

        Future work includes adding a parameter to control the target width and height.

    :param num_output_features:
        Output dimension of the last FC layer in the network which accepts the outputs
        from the convolutional layer. If set to 0, the output will be passed through
        directly to the fusion module.

    :param channel_exp_base:
        Which power of 2 to use in order to set the number of channels in the network.
        For example, setting ``channel_exp_base=3`` means that 2**3=8 channels will be
        used.

    :param first_channel_expansion:
        Factor to extend the first layer channels.

    :param kernel_width:
        Base kernel width of the convolutions.

    :param first_kernel_expansion_width:
        Factor to extend the first kernel's width. The result of the multiplication
        will be rounded to the nearest integer.

    :param down_stride_width:
        Down stride of the convolutional layers along the width.

    :param first_stride_expansion_width:
        Factor to extend the first layer stride along the width. The result of the
        multiplication will be rounded to the nearest integer.

    :param dilation_factor_width:
        Base dilation factor of the convolutions along the width in the network.

    :param kernel_height:
        Base kernel height of the convolutions.

    :param first_kernel_expansion_height:
        Factor to extend the first kernel's height. The result of the multiplication
        will be rounded to the nearest integer.

    :param down_stride_height:
        Down stride of the convolutional layers along the height.

    :param first_stride_expansion_height:
        Factor to extend the first layer stride along the height.
        The result of the multiplication will be rounded to the nearest integer.

    :param dilation_factor_height:
        Base dilation factor of the convolutions along the height in the network.

    :param allow_first_conv_size_reduction:
        If set to False, will not allow the first convolutional layer to reduce the
        size of the input. Setting this is true if you want to ensure that the first
        convolutional layer reduces the size of the input, for example when the input
        is very large, and we want to compress it early.

    :param cutoff:
        If the *resulting* dimension of width * height of adding a successive block
        is less than this value, will stop adding residual blocks to the
        model in the automated case (i.e., if the layers argument is not specified).

    :param rb_do:
        Dropout in the convolutional residual blocks.

    :param stochastic_depth_p:
        Probability of dropping input.

    :param attention_inclusion_cutoff:
        If the dimension of width * height is less than this value, attention will be
        included in the model across channels and width * height as embedding dimension
        after that point (with the channels representing the length of the sequence).

    :param l1:
        L1 regularization to apply to the first layer.
    """

    layers: Union[None, List[int]] = None

    num_output_features: int = 0

    channel_exp_base: int = 2
    first_channel_expansion: int = 1

    kernel_width: int = 12
    first_kernel_expansion_width: float = 1.0
    down_stride_width: int = 4
    first_stride_expansion_width: float = 1.0
    dilation_factor_width: int = 1

    kernel_height: int = 4
    first_kernel_expansion_height: float = 1.0
    down_stride_height: int = 1
    first_stride_expansion_height: float = 1.0
    dilation_factor_height: int = 1

    allow_first_conv_size_reduction: bool = True

    down_sample_every_n_blocks: Optional[int] = 2

    cutoff: int = 32

    rb_do: float = 0.00

    stochastic_depth_p: float = 0.00

    attention_inclusion_cutoff: int = 256

    l1: float = 0.00


def _validate_cnn_config(model_config: CNNModelConfig) -> None:
    mc = model_config

    if mc.down_stride_width > mc.kernel_width:
        raise ValueError(
            f"Down stride width {mc.down_stride_width}"
            f" is greater than kernel width {mc.kernel_width}. "
            f"This is currently not supported."
        )

    if mc.down_stride_height > mc.kernel_height:
        raise ValueError(
            f"Down stride height {mc.down_stride_height} "
            f" is greater than kernel height {mc.kernel_height}."
            f"This is currently not supported."
        )

    first_stride_w = int(round(mc.down_stride_width * mc.first_stride_expansion_width))
    first_kernel_w = int(round(mc.kernel_width * mc.first_kernel_expansion_width))

    first_stride_h = int(
        round(mc.down_stride_height * mc.first_stride_expansion_height)
    )
    first_kernel_h = int(round(mc.kernel_height * mc.first_kernel_expansion_height))

    if first_stride_w > first_kernel_w:
        raise ValueError(
            f"Effective down stride width {first_stride_w}"
            f" (down stride width {mc.down_stride_width}"
            f" times first stride expansion width {mc.first_stride_expansion_width})"
            f" is greater than effective kernel width {first_kernel_w}"
            f" (kernel width {mc.kernel_width}"
            f" times first kernel expansion width {mc.first_kernel_expansion_width})."
            f"This is currently not supported."
        )

    if first_stride_h > first_kernel_h:
        raise ValueError(
            f"Effective down stride height {first_stride_h} "
            f" (down stride height {mc.down_stride_height} "
            f"times first stride expansion height {mc.first_stride_expansion_height})"
            f" is greater than effective kernel height {first_kernel_h}"
            f" (kernel height {mc.kernel_height}"
            f" times first kernel expansion height {mc.first_kernel_expansion_height})."
            f"This is currently not supported."
        )

    positive_int_params = [
        "first_channel_expansion",
        "kernel_width",
        "first_kernel_expansion_width",
        "down_stride_width",
        "first_stride_expansion_width",
        "dilation_factor_width",
        "kernel_height",
        "first_kernel_expansion_height",
        "down_stride_height",
        "first_stride_expansion_height",
        "dilation_factor_height",
        "cutoff",
    ]
    for param in positive_int_params:
        if getattr(mc, param) <= 0:
            raise ValueError(f"{param} must be a positive integer.")

    float_params = ["rb_do", "stochastic_depth_p"]
    for param in float_params:
        value = getattr(mc, param)
        if not (0 <= value <= 1):
            raise ValueError(f"{param} must be in the range [0, 1].")

    if mc.l1 < 0:
        raise ValueError("l1 must be non-negative.")

    if mc.layers is not None:
        if not all(isinstance(layer, int) and layer > 0 for layer in mc.layers):
            raise ValueError(
                "layers must be a list of positive integers "
                "(or None for automatic setup)."
            )


class CNNModel(nn.Module):
    def __init__(
        self,
        model_config: CNNModelConfig,
        data_dimensions: "DataDimensions",
    ):
        super().__init__()

        _validate_cnn_config(model_config=model_config)

        self.model_config = model_config
        self.data_dimensions = data_dimensions

        self.pos_representation = GeneralPositionalEmbedding(
            embedding_dim=self.data_dimensions.height,
            max_length=self.data_dimensions.width,
            dropout=0.0,
        )

        self.conv = nn.Sequential(
            *_make_conv_layers(
                residual_blocks=self.residual_blocks,
                cnn_model_configuration=self.model_config,
                data_dimensions=self.data_dimensions,
            )
        )

        size_func = calc_size_after_conv_sequence
        size_after_conv_w, size_after_conv_h = size_func(
            input_width=self.data_dimensions.width,
            input_height=self.data_dimensions.height,
            conv_sequence=self.conv,
        )
        self.data_size_after_conv = size_after_conv_w * size_after_conv_h

        self.no_out_channels = self.conv[-1].out_channels

        self.output_shape = (self.no_out_channels, size_after_conv_h, size_after_conv_w)

        self.final_layer = (
            nn.Sequential(
                nn.Flatten(start_dim=1),
                nn.Linear(
                    self.no_out_channels * self.data_size_after_conv,
                    model_config.num_output_features,
                ),
            )
            if model_config.num_output_features > 0
            else nn.Identity()
        )

        self._init_weights()

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.conv[0].conv_1.weight

    @property
    def num_out_features(self) -> int:
        if self.model_config.num_output_features > 0:
            return self.model_config.num_output_features
        else:
            return self.no_out_channels * self.data_size_after_conv

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # Swish slope is roughly 0.5 around 0
                nn.init.kaiming_normal_(tensor=m.weight, a=0.5, mode="fan_out")

    @property
    def residual_blocks(self) -> List[int]:
        mc = self.model_config
        if not mc.layers:
            stride_w = mc.down_stride_width
            stride_h = mc.down_stride_height
            down_every = mc.down_sample_every_n_blocks

            if stride_w < 2 and stride_h < 2 and not down_every:
                raise ValueError(
                    "At least one of the strides must be greater than 1, "
                    "or down_sample_every_n_blocks must be set "
                    "when automatic residual block calculation is used "
                    "for CNN model creation. Got: "
                    f"down_stride_width: {stride_w}, "
                    f"down_stride_height: {stride_h}, "
                    f" and down_sample_every_n_blocks: {down_every}."
                )

            residual_blocks = auto_find_no_cnn_residual_blocks_needed(
                input_size_w=self.data_dimensions.width,
                kernel_w=mc.kernel_width,
                first_kernel_expansion_w=mc.first_kernel_expansion_width,
                stride_w=mc.down_stride_width,
                first_stride_expansion_w=mc.first_stride_expansion_width,
                dilation_w=mc.dilation_factor_width,
                down_sample_every_n_blocks=mc.down_sample_every_n_blocks,
                input_size_h=self.data_dimensions.height,
                kernel_h=mc.kernel_height,
                first_kernel_expansion_h=mc.first_kernel_expansion_height,
                stride_h=mc.down_stride_height,
                first_stride_expansion_h=mc.first_stride_expansion_height,
                dilation_h=mc.dilation_factor_height,
                cutoff=mc.cutoff,
            )

            logger.debug(
                "No residual blocks specified in CL args, using input "
                "%s based on size approximation calculation.",
                residual_blocks,
            )
            return residual_blocks

        assert isinstance(mc.layers, list)
        return mc.layers

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = self.pos_representation(input)
        out = self.conv(out)
        out = self.final_layer(out)

        return out


def _make_conv_layers(
    residual_blocks: List[int],
    cnn_model_configuration: CNNModelConfig,
    data_dimensions: "DataDimensions",
) -> List[nn.Module]:
    mc = cnn_model_configuration

    first_conv_channels = 2**mc.channel_exp_base * mc.first_channel_expansion

    down_stride_w = mc.down_stride_width
    first_conv_kernel_w = int(round(mc.kernel_width * mc.first_kernel_expansion_width))
    first_conv_stride_w = int(round(down_stride_w * mc.first_stride_expansion_width))

    conv_param_suggestion_w = calc_conv_params_needed(
        input_size=data_dimensions.width,
        kernel_size=first_conv_kernel_w,
        stride=first_conv_stride_w,
        dilation=mc.dilation_factor_width,
        allow_non_stride_reduction=mc.allow_first_conv_size_reduction,
    )

    down_stride_h = mc.down_stride_height
    first_conv_kernel_h = int(
        round(mc.kernel_height * mc.first_kernel_expansion_height)
    )
    first_conv_stride_h = int(
        round(mc.down_stride_height * mc.first_stride_expansion_height)
    )

    conv_param_suggestion_h = calc_conv_params_needed(
        input_size=data_dimensions.height,
        kernel_size=first_conv_kernel_h,
        stride=first_conv_stride_h,
        dilation=mc.dilation_factor_height,
        allow_non_stride_reduction=mc.allow_first_conv_size_reduction,
    )

    conv_blocks: list[FirstCNNBlock | CNNResidualBlock | ConvAttentionBlock | nn.Module]
    conv_blocks = [
        FirstCNNBlock(
            in_channels=data_dimensions.channels,
            out_channels=first_conv_channels,
            conv_1_kernel_h=conv_param_suggestion_h.kernel_size,
            conv_1_kernel_w=conv_param_suggestion_w.kernel_size,
            conv_1_padding_w=conv_param_suggestion_w.padding,
            conv_1_padding_h=conv_param_suggestion_h.padding,
            down_stride_w=first_conv_stride_w,
            down_stride_h=first_conv_stride_h,
            dilation_w=conv_param_suggestion_w.dilation,
            dilation_h=conv_param_suggestion_h.dilation,
            rb_do=mc.rb_do,
        )
    ]

    first_width, first_height = calc_size_after_conv_sequence(
        input_width=data_dimensions.width,
        input_height=data_dimensions.height,
        conv_sequence=nn.Sequential(*conv_blocks),
    )
    do_add_attention = _do_add_attention(
        width=first_width,
        height=first_height,
        attention_inclusion_cutoff=mc.attention_inclusion_cutoff,
    )

    if do_add_attention:
        last_block = conv_blocks[-1]
        assert isinstance(last_block, (CNNResidualBlock, FirstCNNBlock))
        cur_attention_block = ConvAttentionBlock(
            channels=last_block.out_channels,
            width=first_width,
            height=first_height,
        )
        conv_blocks.append(cur_attention_block)

    down_every = mc.down_sample_every_n_blocks
    for block_arch_idx, block_arch_blocks in enumerate(residual_blocks):
        for layer in range(block_arch_blocks):
            cur_block = _get_conv_residual_block(
                conv_blocks=conv_blocks,
                layer_arch_idx=block_arch_idx,
                down_stride_w=down_stride_w,
                down_stride_h=down_stride_h,
                cnn_config=cnn_model_configuration,
                data_dimensions=data_dimensions,
            )

            conv_blocks.append(cur_block)

            cur_width, cur_height = calc_size_after_conv_sequence(
                input_width=data_dimensions.width,
                input_height=data_dimensions.height,
                conv_sequence=nn.Sequential(*conv_blocks),
            )

            n_blocks = len([i for i in conv_blocks if isinstance(i, CNNResidualBlock)])
            if down_every and n_blocks % down_every == 0:
                down_block = DownSamplingResidualBlock(
                    in_channels=conv_blocks[-1].out_channels,
                    in_width=cur_width,
                    in_height=cur_height,
                )
                conv_blocks.append(down_block)

                cur_width, cur_height = calc_size_after_conv_sequence(
                    input_width=data_dimensions.width,
                    input_height=data_dimensions.height,
                    conv_sequence=nn.Sequential(*conv_blocks),
                )

            do_add_attention = _do_add_attention(
                width=cur_width,
                height=cur_height,
                attention_inclusion_cutoff=mc.attention_inclusion_cutoff,
            )
            if do_add_attention:
                cur_attention_block = ConvAttentionBlock(
                    channels=cur_block.out_channels,
                    width=cur_width,
                    height=cur_height,
                )
                conv_blocks.append(cur_attention_block)

    return conv_blocks


def _do_add_attention(
    width: int,
    height: int,
    attention_inclusion_cutoff: int,
) -> bool:
    if attention_inclusion_cutoff == 0:
        return False

    if width * height <= attention_inclusion_cutoff:
        return True

    return False


def _get_conv_residual_block(
    conv_blocks: List[nn.Module],
    layer_arch_idx: int,
    down_stride_w: int,
    down_stride_h: int,
    cnn_config: CNNModelConfig,
    data_dimensions: "DataDimensions",
) -> CNNResidualBlock:
    mc = cnn_config

    cur_conv = nn.Sequential(*conv_blocks)
    cur_width, cur_height = calc_size_after_conv_sequence(
        input_width=data_dimensions.width,
        input_height=data_dimensions.height,
        conv_sequence=cur_conv,
    )

    cur_block_number = (
        len([i for i in conv_blocks if isinstance(i, CNNResidualBlock)]) + 1
    )

    cur_dilation_factor_w = _get_cur_dilation(
        dilation_factor=mc.dilation_factor_width,
        size=cur_width,
        block_number=cur_block_number,
        kernel_size=mc.kernel_width,
    )
    conv_param_suggestion_w = calc_conv_params_needed(
        input_size=cur_width,
        kernel_size=mc.kernel_width,
        stride=down_stride_w,
        dilation=cur_dilation_factor_w,
    )

    cur_dilation_factor_h = _get_cur_dilation(
        dilation_factor=mc.dilation_factor_height,
        size=cur_height,
        block_number=cur_block_number,
        kernel_size=mc.kernel_height,
    )
    conv_param_suggestion_h = calc_conv_params_needed(
        input_size=cur_height,
        kernel_size=mc.kernel_height,
        stride=down_stride_h,
        dilation=cur_dilation_factor_h,
    )

    cur_in_channels = conv_blocks[-1].out_channels
    cur_out_channels = 2 ** (mc.channel_exp_base + layer_arch_idx)

    w_stays_same = conv_param_suggestion_h.target_size == cur_height
    h_stays_same = conv_param_suggestion_w.target_size == cur_width
    channels_stays_same = cur_in_channels == cur_out_channels

    need_identity_downsample = True
    if w_stays_same and h_stays_same and channels_stays_same:
        need_identity_downsample = False

    cur_layer = CNNResidualBlock(
        in_channels=cur_in_channels,
        out_channels=cur_out_channels,
        conv_1_kernel_w=conv_param_suggestion_w.kernel_size,
        conv_1_padding_w=conv_param_suggestion_w.padding,
        down_stride_w=conv_param_suggestion_w.stride,
        dilation_w=conv_param_suggestion_w.dilation,
        conv_1_kernel_h=conv_param_suggestion_h.kernel_size,
        conv_1_padding_h=conv_param_suggestion_h.padding,
        down_stride_h=conv_param_suggestion_h.stride,
        dilation_h=conv_param_suggestion_h.dilation,
        full_preact=True if len(conv_blocks) == 1 else False,
        rb_do=mc.rb_do,
        stochastic_depth_p=mc.stochastic_depth_p,
        conv_downsample_identity=need_identity_downsample,
    )

    return cur_layer


def _get_cur_dilation(
    dilation_factor: int, size: int, block_number: int, kernel_size: int
):
    """
    Note that block_number refers to the number of full residual blocks
    (excluding the first one).
    """
    if size == 1 or kernel_size == 1:
        return 1

    dilation = dilation_factor**block_number

    max_dilation = max(1, (size - 1) // (kernel_size - 1))
    while dilation > max_dilation:
        dilation = dilation // dilation_factor

    return dilation


def auto_find_no_cnn_residual_blocks_needed(
    input_size_w: int,
    kernel_w: int,
    first_kernel_expansion_w: float,
    stride_w: int,
    first_stride_expansion_w: float,
    dilation_w: int,
    down_sample_every_n_blocks: Optional[int],
    input_size_h: int,
    kernel_h: int,
    first_kernel_expansion_h: float,
    stride_h: int,
    first_stride_expansion_h: float,
    dilation_h: int,
    cutoff: int,
) -> List[int]:
    """
    Used in order to calculate / set up residual blocks specifications as a list
    automatically when they are not passed in as CL args, based on the minimum
    size after the residual block convolutions.

    We have 2 residual_blocks per channel depth until we have a total of 8 blocks,
    then the rest is put in the third depth index (following resnet convention).

    That is with a base channel depth of 32, we have these depths in the list:
    [32, 64, 128, 256].

    Examples
    ------
    3 blocks --> [2, 1]
    7 blocks --> [2, 2, 2, 1]
    10 blocks --> [2, 2, 4, 2]
    """

    if (
        (stride_w == 1 and input_size_w > cutoff)
        or (stride_h == 1 and input_size_h > cutoff)
    ) and not down_sample_every_n_blocks:
        err_dim = "width" if stride_w == 1 else "height"
        logger.warning(
            f"With stride=1, the {err_dim} size "
            f"({input_size_w if err_dim == 'width' else input_size_h}) "
            f"cannot be larger than the cutoff ({cutoff}). "
            f"This would result in an infinite loop."
            f"Will stop when no more reduction is possible,"
            f"despite not reaching the cutoff."
        )

    first_kernel_w = int(round(kernel_w * first_kernel_expansion_w))
    first_stride_w = int(round(stride_w * first_stride_expansion_w))

    conv_param_suggestion_w_first = calc_conv_params_needed(
        input_size=input_size_w,
        kernel_size=first_kernel_w,
        stride=first_stride_w,
        dilation=dilation_w,
    )

    first_kernel_h = int(round(kernel_h * first_kernel_expansion_h))
    first_stride_h = int(round(stride_h * first_stride_expansion_h))

    conv_param_suggestion_h_first = calc_conv_params_needed(
        input_size=input_size_h,
        kernel_size=first_kernel_h,
        stride=first_stride_h,
        dilation=dilation_h,
    )

    cur_size_w = conv_output_formula(
        input_size=input_size_w,
        kernel_size=conv_param_suggestion_w_first.kernel_size,
        stride=first_stride_w,
        dilation=dilation_w,
        padding=conv_param_suggestion_w_first.padding,
    )
    cur_size_h = conv_output_formula(
        input_size=input_size_h,
        kernel_size=conv_param_suggestion_h_first.kernel_size,
        stride=first_stride_h,
        dilation=dilation_h,
        padding=conv_param_suggestion_h_first.padding,
    )

    residual_blocks = [0] * 4
    down_every = down_sample_every_n_blocks
    while True:
        if down_every is not None and sum(residual_blocks) % down_every == 0:
            cur_size_w = max(1, cur_size_w // 2)
            cur_size_h = max(1, cur_size_h // 2)

        conv_param_suggestion_w = calc_conv_params_needed(
            input_size=cur_size_w,
            kernel_size=kernel_w,
            stride=stride_w,
            dilation=dilation_w,
        )
        cur_size_w_next = conv_output_formula(
            input_size=cur_size_w,
            kernel_size=conv_param_suggestion_w.kernel_size,
            stride=conv_param_suggestion_w.stride,
            dilation=conv_param_suggestion_w.dilation,
            padding=conv_param_suggestion_w.padding,
        )

        conv_param_suggestion_h = calc_conv_params_needed(
            input_size=cur_size_h,
            kernel_size=kernel_h,
            stride=stride_h,
            dilation=dilation_h,
        )
        cur_size_h_next = conv_output_formula(
            input_size=cur_size_h,
            kernel_size=conv_param_suggestion_h.kernel_size,
            stride=conv_param_suggestion_h.stride,
            dilation=conv_param_suggestion_h.dilation,
            padding=conv_param_suggestion_h.padding,
        )

        cannot_reduce_more = (
            cur_size_w == cur_size_w_next and cur_size_h == cur_size_h_next
        )

        w_kernel_larger = conv_param_suggestion_w.kernel_size > cur_size_w
        h_kernel_larger = conv_param_suggestion_h.kernel_size > cur_size_h
        kernel_too_large = w_kernel_larger or h_kernel_larger
        if (
            cur_size_w_next * cur_size_h_next < cutoff
            or (cannot_reduce_more and down_every is None)
            or kernel_too_large
        ):
            if cannot_reduce_more:
                logger.warning(
                    f"Could not reduce size more, "
                    f"despite not reaching the cutoff ({cutoff})."
                )

            break

        else:
            cur_no_blocks = sum(residual_blocks)

            if cur_no_blocks >= 8:
                residual_blocks[2] += 1
            else:
                cur_index = cur_no_blocks // 2
                residual_blocks[cur_index] += 1

            cur_size_w, cur_size_h = cur_size_w_next, cur_size_h_next

    return [i for i in residual_blocks if i != 0]


class GeneralPositionalEmbedding(nn.Module):
    def __init__(
        self,
        embedding_dim: int,
        max_length: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.embedding_dim = embedding_dim
        self.max_length = max_length
        self.dropout = nn.Dropout(p=dropout)

        self.embedding = torch.nn.Parameter(
            data=torch.zeros(1, self.embedding_dim, self.max_length),
            requires_grad=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.embedding
        return self.dropout(x)


def calc_size_after_conv_sequence(
    input_width: int,
    input_height: int,
    conv_sequence: nn.Sequential,
) -> Tuple[int, int]:
    current_width = input_width
    current_height = input_height
    for block_index, block in enumerate(conv_sequence):
        conv_operations = [i for i in vars(block)["_modules"] if i.find("conv") != -1]

        for operation_index, operation in enumerate(conv_operations):
            conv_layer = vars(block)["_modules"][operation]

            padded_height = current_height + 2 * conv_layer.padding[0]
            padded_width = current_width + 2 * conv_layer.padding[1]
            if any(
                k > s
                for k, s in zip(conv_layer.kernel_size, (padded_height, padded_width))
            ):
                raise ValueError(
                    f"Kernel size of layer "
                    f"{block_index}.{operation_index} ({operation}) "
                    f"exceeds padded input size in one or more dimensions. "
                    f"Original input size (hxw): {current_height}x{current_width} "
                    f"(this is likely the source of the problem, especially if "
                    f"error layer is conv_1). "
                    f"Padded input size became (hxw): {padded_height}x{padded_width}. "
                    f"Kernel size: {conv_layer.kernel_size}. "
                    "Please adjust the kernel size to ensure the it "
                    "does not exceed the padded input size for each dimension."
                )

            new_width = _calc_layer_output_size_for_axis(
                size=current_width, layer=conv_layer, axis=1
            )
            new_height = _calc_layer_output_size_for_axis(
                size=current_height, layer=conv_layer, axis=0
            )

            if int(new_width) == 0 or int(new_height) == 0:
                kernel_size = conv_layer.kernel_size
                stride = conv_layer.stride
                padding = conv_layer.padding
                dilation = conv_layer.dilation

                raise ValueError(
                    f"Calculated size after convolution sequence is 0 for layer "
                    f"{block_index}.{operation_index} ({operation}). "
                    f"Input size (hxw): {current_height}x{current_width}. "
                    f"Convolution parameters: kernel size = {kernel_size}, "
                    f"stride = {stride}, padding = {padding}, dilation = {dilation}. "
                    "Please adjust these parameters to ensure they are appropriate "
                    "for the input size."
                )

            current_width, current_height = new_width, new_height

    return int(current_width), int(current_height)


def _calc_layer_output_size_for_axis(
    size: int, layer: nn.Conv1d | nn.Conv2d | nn.Conv3d, axis: int
):
    kernel_size = layer.kernel_size[axis]
    padding = layer.padding[axis]
    assert isinstance(padding, int)
    stride = layer.stride[axis]
    dilation = layer.dilation[axis]

    output_size = conv_output_formula(
        input_size=size,
        padding=padding,
        dilation=dilation,
        kernel_size=kernel_size,
        stride=stride,
    )

    return output_size


@dataclass()
class ConvParamSuggestion:
    kernel_size: int
    target_size: int
    stride: int
    dilation: int
    padding: int


def calc_conv_params_needed(
    input_size: int,
    kernel_size: int,
    stride: int,
    dilation: int,
    allow_non_stride_reduction: bool = False,
) -> "ConvParamSuggestion":
    if input_size < 0:
        raise ValueError("Got negative size for input width: %d", input_size)

    if stride == 1 and not allow_non_stride_reduction:
        target_size = input_size
    else:
        target_size = conv_output_formula(
            input_size=input_size,
            padding=0,
            dilation=dilation,
            kernel_size=kernel_size,
            stride=stride,
        )

    target_size = max(1, target_size)

    param_suggestions = _get_conv_param_suggestion_iterator(
        target_size=target_size,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
    )

    solutions = []

    for s in param_suggestions:
        padding = solve_for_padding(
            input_size=input_size,
            target_size=s.target_size,
            dilation=s.dilation,
            stride=s.stride,
            kernel_size=s.kernel_size,
        )

        if padding is not None:
            assert isinstance(padding, int)
            cur_solution = ConvParamSuggestion(
                kernel_size=s.kernel_size,
                target_size=s.target_size,
                stride=s.stride,
                dilation=s.dilation,
                padding=padding,
            )
            solutions.append(cur_solution)

    if len(solutions) == 0:
        raise AssertionError(
            f"Could not find a solution for padding with the supplied conv "
            f"parameters: input size = {input_size}, kernel size = {kernel_size}, "
            f"stride = {stride}, dilation = {dilation}. "
        )

    best_solution = _choose_best_solution(
        solutions=solutions,
        input_target_size=target_size,
        input_kernel_size=kernel_size,
        input_stride=stride,
        input_dilation=dilation,
    )

    return best_solution


def _choose_best_solution(
    solutions: List[ConvParamSuggestion],
    input_target_size: int,
    input_kernel_size: int,
    input_stride: int,
    input_dilation: int,
) -> ConvParamSuggestion:
    def _calculate_distance(
        solution: ConvParamSuggestion,
    ) -> Tuple[int, Tuple[int, ...]]:
        """
        We have the second returned value as a tuple so that if >1 solutions
        have the same distance, we get a consistent ordering.

        We have *10 there to prioritize maintaining the target size
        over other parameters.
        """
        kernel_size_diff = abs(solution.kernel_size - input_kernel_size)
        stride_diff = abs(solution.stride - input_stride)
        dilation_diff = abs(solution.dilation - input_dilation)
        padding_diff = abs(solution.padding)
        target_size_diff = abs(solution.target_size - input_target_size) * 10

        distance = (
            kernel_size_diff
            + stride_diff
            + dilation_diff
            + padding_diff
            + target_size_diff
        )

        values_tuple = (
            solution.kernel_size,
            solution.stride,
            solution.dilation,
            solution.padding,
            solution.target_size,
        )

        return distance, values_tuple

    sorted_solutions = sorted(solutions, key=_calculate_distance)

    return sorted_solutions[0]


def _get_conv_param_suggestion_iterator(
    target_size: int, kernel_size: int, stride: int, dilation: int
) -> Iterator[ConvParamSuggestion]:
    def _get_range(base: int) -> List[int]:
        return [sug for sug in [base, base + 1, base - 1] if sug > 0]

    for k_size in _get_range(base=kernel_size):
        for t_size in _get_range(base=target_size):
            for s_size in _get_range(base=stride):
                for d_size in _get_range(base=dilation):
                    yield ConvParamSuggestion(
                        kernel_size=k_size,
                        target_size=t_size,
                        stride=s_size,
                        dilation=d_size,
                        padding=0,
                    )


def conv_output_formula(
    input_size: int, padding: int, dilation: int, kernel_size: int, stride: int
) -> int:
    out_size = (
        input_size + 2 * padding - dilation * (kernel_size - 1) - 1
    ) // stride + 1
    return out_size


def solve_for_padding(
    input_size: int, target_size: int, dilation: int, stride: int, kernel_size: int
) -> Union[int, None]:
    p = Symbol("p", integer=True, nonnegative=True)
    padding = solve(
        ((input_size + (2 * p) - dilation * (kernel_size - 1) - 1) / stride + 1)
        - target_size,
        p,
    )

    if len(padding) > 0:
        return int(padding[0])

    return None



from typing import Any, Dict, Literal, Tuple, Union

import torch

from eir.predict_modules.predict_tabular_input_setup import (
    ComputedPredictTabularInputInfo,
)
from eir.serve_modules.serve_schemas import ComputedServeTabularInputInfo
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.input_setup_modules.setup_tabular import ComputedTabularInputInfo
from eir.setup.output_setup import al_output_objects_as_dict
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
)
from eir.setup.output_setup_modules.survival_output_setup import (
    ComputedSurvivalOutputInfo,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
)
from eir.setup.schemas import (
    ImageInputDataConfig,
    ImageOutputTypeConfig,
    SurvivalOutputTypeConfig,
    TabularInputDataConfig,
    TabularOutputTypeConfig,
)

al_fill_values = dict[str, bool | int | float | dict[str, int | float]]


def impute_missing_modalities_wrapper(
    inputs_values: dict[str, Any], inputs_objects: "al_input_objects_as_dict"
) -> Dict[str, torch.Tensor]:
    impute_dtypes = _get_default_impute_dtypes(inputs_objects=inputs_objects)
    impute_fill_values = _get_default_impute_fill_values(inputs_objects=inputs_objects)
    inputs_imputed = impute_missing_modalities(
        inputs_values=inputs_values,
        inputs_objects=inputs_objects,
        fill_values=impute_fill_values,
        dtypes=impute_dtypes,
    )

    return inputs_imputed


def impute_missing_modalities(
    inputs_values: dict[str, Any],
    inputs_objects: al_input_objects_as_dict,
    fill_values: al_fill_values,
    dtypes: dict[str, Any],
) -> dict[str, torch.Tensor]:
    for input_name, input_object in inputs_objects.items():
        input_type = input_object.input_config.input_info.input_type

        if input_name not in inputs_values:
            fill_value = fill_values[input_name]
            dtype = dtypes[input_name]

            shape: Tuple[int, ...]
            approach: Literal["constant", "random"]
            match input_object:
                case ComputedOmicsInputInfo():
                    assert input_type == "omics"
                    dimensions = input_object.data_dimensions
                    shape = dimensions.channels, dimensions.height, dimensions.width
                    approach = "constant"

                case ComputedSequenceInputInfo() | ComputedBytesInputInfo():
                    assert input_type in ("sequence", "bytes")
                    max_length = input_object.computed_max_length
                    shape = (max_length,)
                    approach = "constant"

                case ComputedImageInputInfo():
                    assert input_type == "image"
                    input_type_info = input_object.input_config.input_type_info
                    assert isinstance(input_type_info, ImageInputDataConfig)
                    size = input_type_info.size
                    if len(size) == 1:
                        size = [size[0], size[0]]

                    num_channels = input_object.num_channels
                    shape = (num_channels, *size)
                    approach = "random"

                case (
                    ComputedTabularInputInfo()
                    | ComputedPredictTabularInputInfo()
                    | ComputedServeTabularInputInfo()
                ):
                    assert input_type == "tabular"
                    inputs_values[input_name] = fill_value
                    continue

                case ComputedArrayInputInfo():
                    assert input_type == "array"
                    shape = input_object.data_dimensions.full_shape()
                    approach = "random"

                case _:
                    raise ValueError(f"Unrecognized input type {input_type}")

            imputed_tensor = impute_single_missing_modality(
                shape=shape,
                fill_value=fill_value,
                dtype=dtype,
                approach=approach,
            )
            inputs_values[input_name] = imputed_tensor

    return inputs_values


def impute_single_missing_modality(
    shape: Tuple[int, ...],
    fill_value: Any,
    dtype: Any,
    approach: Literal["constant", "random"],
) -> torch.Tensor:
    match approach:
        case "constant":
            imputed_tensor = torch.empty(shape, dtype=dtype).fill_(fill_value)
        case "random":
            imputed_tensor = torch.empty(shape, dtype=dtype).normal_()
        case _:
            raise ValueError(f"Unrecognized approach {approach}")

    return imputed_tensor


def _get_default_impute_fill_values(
    inputs_objects: "al_input_objects_as_dict",
) -> al_fill_values:
    fill_values: al_fill_values = {}
    for input_name, input_object in inputs_objects.items():
        match input_object:
            case ComputedOmicsInputInfo():
                fill_values[input_name] = False

            case (
                ComputedTabularInputInfo()
                | ComputedPredictTabularInputInfo()
                | ComputedServeTabularInputInfo()
            ):
                fill_values[input_name] = _build_tabular_fill_value(
                    input_object=input_object
                )

            case (
                ComputedSequenceInputInfo()
                | ComputedBytesInputInfo()
                | ComputedImageInputInfo()
                | ComputedArrayInputInfo()
            ):
                fill_values[input_name] = 0.0
            case _:
                raise ValueError(
                    f"Unrecognized input type"
                    f" {input_object.input_config.input_info.input_type}"
                )

    return fill_values


def _build_tabular_fill_value(
    input_object: Union[
        "ComputedTabularInputInfo",
        "ComputedPredictTabularInputInfo",
        "ComputedServeTabularInputInfo",
    ]
) -> dict[str, int | float]:
    fill_value = {}
    transformers = input_object.labels.label_transformers

    input_type_info = input_object.input_config.input_type_info
    assert isinstance(input_type_info, TabularInputDataConfig)

    cat_columns = input_type_info.input_cat_columns
    for cat_column in cat_columns:
        cur_label_encoder = transformers[cat_column]
        fill_value[cat_column] = cur_label_encoder.transform(["__NULL__"]).item()

    con_columns = input_type_info.input_con_columns
    for con_column in con_columns:
        # already assumes mean normalized values, so fill with 0.0
        fill_value[con_column] = 0.0

    return fill_value


def _get_default_impute_dtypes(inputs_objects: "al_input_objects_as_dict"):
    dtypes = {}
    for input_name, input_object in inputs_objects.items():
        input_type = input_object.input_config.input_info.input_type
        if input_type == "omics":
            dtypes[input_name] = torch.bool
        elif input_type in ("sequence", "bytes"):
            dtypes[input_name] = torch.long
        else:
            dtypes[input_name] = torch.float

    return dtypes


def impute_missing_output_modalities_wrapper(
    outputs_values: dict[str, Any], output_objects: "al_output_objects_as_dict"
) -> dict[str, dict[str, torch.Tensor | int | float]]:
    impute_dtypes = _get_default_output_impute_dtypes(outputs_objects=output_objects)
    impute_fill_values = _get_default_output_impute_fill_values(
        outputs_objects=output_objects
    )
    outputs_imputed_modalities = impute_missing_output_modalities(
        outputs_values=outputs_values,
        outputs_objects=output_objects,
        fill_values=impute_fill_values,
        dtypes=impute_dtypes,
    )

    outputs_imputed = impute_partially_missing_output_modalities(
        outputs_values=outputs_imputed_modalities,
        output_objects=output_objects,
    )

    return outputs_imputed


def impute_partially_missing_output_modalities(
    outputs_values: dict[str, Any],
    output_objects: "al_output_objects_as_dict",
) -> dict[str, dict[str, torch.Tensor | int | float]]:
    for output_name, output_object in output_objects.items():
        match output_object:
            case ComputedTabularOutputInfo():
                cur_output_value = outputs_values[output_name]
                output_type_info = output_object.output_config.output_type_info

                assert isinstance(output_type_info, TabularOutputTypeConfig)

                cat_columns = set(output_type_info.target_cat_columns)
                con_columns = set(output_type_info.target_con_columns)
                output_columns = cat_columns.union(con_columns)

                for output_column in output_columns:
                    if output_column not in cur_output_value:
                        cur_output_value[output_column] = torch.nan

            case ComputedSurvivalOutputInfo():
                cur_output_value = outputs_values[output_name]
                output_type_info = output_object.output_config.output_type_info

                assert isinstance(output_type_info, SurvivalOutputTypeConfig)
                event_column = output_type_info.event_column
                time_column = output_type_info.time_column

                if event_column not in cur_output_value:
                    cur_output_value[event_column] = torch.nan
                if time_column not in cur_output_value:
                    cur_output_value[time_column] = torch.nan

                outputs_values[output_name] = cur_output_value

    return outputs_values


def impute_missing_output_modalities(
    outputs_values: dict[str, Any],
    outputs_objects: "al_output_objects_as_dict",
    fill_values: al_fill_values,
    dtypes: dict[str, Any],
) -> dict[str, dict[str, torch.Tensor | int | float]]:
    """
    Note that ultimately we never want to use these values for anything e.g. in the
    loss calculation, but rather just skip them completely. However, we do need
    the shapes to match for collation purposes.

    Note we skip imputing sequence outputs as they are handled on the fly
    separately based on the input.
    """
    for output_name, output_object in outputs_objects.items():
        output_type = output_object.output_config.output_info.output_type
        if output_name not in outputs_values:
            fill_value = fill_values[output_name]
            dtype = dtypes[output_name]

            shape: Tuple[int, ...]
            approach: Literal["constant", "random"]
            match output_object:
                case ComputedSequenceOutputInfo():
                    continue

                case ComputedArrayOutputInfo():
                    assert output_type == "array"
                    shape = output_object.data_dimensions.full_shape()
                    approach = "constant"

                case ComputedTabularOutputInfo() | ComputedSurvivalOutputInfo():
                    assert output_type in ("tabular", "survival")
                    outputs_values[output_name] = fill_value
                    continue

                case ComputedImageOutputInfo():
                    assert output_type == "image"
                    output_type_info = output_object.output_config.output_type_info
                    assert isinstance(output_type_info, ImageOutputTypeConfig)
                    size = output_type_info.size
                    if len(size) == 1:
                        size = [size[0], size[0]]

                    num_channels = output_object.num_channels
                    shape = (num_channels, *size)
                    approach = "random"

                case _:
                    raise ValueError(f"Unrecognized output type {output_type}")

            imputed_tensor = impute_single_missing_modality(
                shape=shape,
                fill_value=fill_value,
                dtype=dtype,
                approach=approach,
            )
            outputs_values[output_name] = {output_name: imputed_tensor}

    return outputs_values


def _get_default_output_impute_dtypes(
    outputs_objects: al_output_objects_as_dict,
) -> dict[str, Any]:
    dtypes = {}
    for output_name, output_object in outputs_objects.items():
        match output_object:
            case ComputedTabularOutputInfo() | ComputedSurvivalOutputInfo():
                dtypes[output_name] = torch.float
            case ComputedSequenceOutputInfo():
                dtypes[output_name] = torch.long
            case ComputedArrayOutputInfo() | ComputedImageOutputInfo():
                dtypes[output_name] = torch.float
            case _:
                raise ValueError(
                    f"Unrecognized output type"
                    f" {output_object.output_config.output_info.output_type}"
                )

    return dtypes


def _get_default_output_impute_fill_values(
    outputs_objects: al_output_objects_as_dict,
) -> al_fill_values:
    fill_values: al_fill_values = {}
    for output_name, output_object in outputs_objects.items():
        match output_object:
            case ComputedTabularOutputInfo():
                fill_values[output_name] = _build_tabular_output_fill_value(
                    output_object=output_object
                )
            case ComputedSequenceOutputInfo():
                fill_values[output_name] = 0
            case ComputedArrayOutputInfo() | ComputedImageOutputInfo():
                fill_values[output_name] = torch.nan
            case ComputedSurvivalOutputInfo():
                fill_values[output_name] = _build_survival_output_fill_value(
                    output_object=output_object
                )
            case _:
                raise ValueError(
                    f"Unrecognized output type"
                    f" {output_object.output_config.output_info.output_type}"
                )

    return fill_values


def _build_tabular_output_fill_value(
    output_object: Union["ComputedTabularOutputInfo"],
) -> dict[str, int | float]:
    fill_value: dict[str, int | float] = {}

    output_type_info = output_object.output_config.output_type_info
    assert isinstance(output_type_info, TabularOutputTypeConfig)

    cat_columns = output_type_info.target_cat_columns
    for cat_column in cat_columns:
        fill_value[cat_column] = torch.nan

    con_columns = output_type_info.target_con_columns
    for con_column in con_columns:
        fill_value[con_column] = torch.nan

    return fill_value


def _build_survival_output_fill_value(
    output_object: Union["ComputedSurvivalOutputInfo"],
) -> dict[str, int | float]:
    fill_value: dict[str, int | float] = {}

    output_type_info = output_object.output_config.output_type_info
    assert isinstance(output_type_info, SurvivalOutputTypeConfig)

    event_column = output_type_info.event_column
    time_column = output_type_info.time_column

    fill_value[event_column] = torch.nan
    fill_value[time_column] = torch.nan

    return fill_value
